/* ** error.c ************************************ */
/* Error routins for tuitated. */
/* *********************************************** */

/* -- Include File ------------------------------- */
#include <config.h>
#include <stdio.h>
#include <string.h>
#ifndef STDC_HEADERS
#include <varargs.h>
#else
#include <stdarg.h>
#endif
#include <sys/types.h>
#include <unistd.h>
#include <syslog.h>
#include <time.h>
#include "tuitated.h"
#include "struct.h"
#include "gzlib/zlib.h"

/* -- Global Data -------------------------------- */
extern int InDebugMode;
extern int InLogMode;

extern int isAccessLogFileSpecified;
extern int isServiceLogFileSpecified;
extern int isErrorLogFileSpecified;

extern char AccessLogFile[512];
extern char ServiceLogFile[512];
extern char ErrorLogFile[512];


/* -- Protype Declerations  ---------------------- */
int FileOut( char *filename, char *string );

/* ** Func Err *********************************** */
/* Function to log System Error message. */
void Err( char *format, ... )
{
/* -- Local Data --------------------------------- */
    va_list ap;
    time_t t;
    char out_buf[512];
    char work_buf[512];
/* -- Operation ---------------------------------- */

    /* Print Version */
    strcpy( out_buf, VERSION );
    strcat( out_buf, " " );
	
    /* Print Data & Time */
    time( &t );
    strcat( out_buf, ctime( &t ) );
    out_buf[strlen(out_buf) - 1] = '\0';

    strcat( out_buf, " : " );

    /* Print Process ID */
    sprintf( work_buf, "%d", (int)getpid() );
    strcat( out_buf, work_buf );

    strcat( out_buf, " : " );

    /* Print Message */
#ifndef STDC_HEADERS
    va_start( ap );
#else
    va_start( ap, format );
#endif
    vsprintf( work_buf, format, ap );
    va_end( ap );

    strcat( out_buf, work_buf );

    /* Print Return Code */
    strcat( out_buf, "\n" );

    if( InDebugMode ) {
        printf( "%s", out_buf );
    } else if( isErrorLogFileSpecified ) {
        FileOut( ErrorLogFile, out_buf );
    }
} /* Err */

/* ** Func AccessLog *********************************** */
/* Function to log access message. */
void AccessLog( char *format, ... )
{
/* -- Local Data --------------------------------- */
    va_list ap;
    time_t t;
    char out_buf[CLIENT_BUFSIZE + 512];
    char work_buf[CLIENT_BUFSIZE + 512];
/* -- Operation ---------------------------------- */

    /* Print Data & Time */
    time( &t );
    strcpy( out_buf, ctime( &t ) );
    out_buf[strlen(out_buf) - 1] = '\0';

    strcat( out_buf, " : " );

    /* Print Message */
#ifndef STDC_HEADERS
    va_start( ap );
#else
    va_start( ap, format );
#endif
    vsprintf( work_buf, format, ap );
    va_end( ap );

    strcat( out_buf, work_buf );

    /* Print Return Code */
    strcat( out_buf, "\n" );

    if( InDebugMode ) {
        printf( "%s", out_buf );
    } else if( isAccessLogFileSpecified ) {
        FileOut( AccessLogFile, out_buf );
    }
} /* AccessLog */


/* ** Func ServiceLog *********************************** */
/* Function to log service. */
void ServiceLog( char *format, ... )
{
/* -- Local Data --------------------------------- */
    va_list ap;
    time_t t;
    char out_buf[512];
    char work_buf[512];
/* -- Operation ---------------------------------- */

    /* Print Data & Time */
    time( &t );
    strcpy( out_buf, ctime( &t ) );
    out_buf[strlen(out_buf) - 1] = '\0';

    strcat( out_buf, " : " );

    /* Print Message */
#ifndef STDC_HEADERS
    va_start( ap );
#else
    va_start( ap, format );
#endif
    vsprintf( work_buf, format, ap );
    va_end( ap );

    strcat( out_buf, work_buf );

    /* Print Return Code */
    strcat( out_buf, "\n" );

    if( InDebugMode ) {
        printf( "%s", out_buf );
    } else if( isServiceLogFileSpecified ) {
        FileOut( ServiceLogFile, out_buf );
    }
} /* ServiceLog */


/* ** Func GZip ********************************** */
/*
 * Commpress given file with gzip command.ee
 * At the end of the name of the file,
 * ".gz" will be added by gzip command.
 */
int GZip( char *filename )
{
/* -- Local Data --------------------------------- */
    FILE *in;
    gzFile out;
    char gzfilename[512];
    char buf[4096];
    int len;
/* -- Operation ---------------------------------- */
    strcpy( gzfilename, filename );
    strcat( gzfilename, ".gz" );
    in = fopen( filename, "r" );
    if( in == NULL ) {
        Err( "Cannot Open %s.", filename );
        return 1;
    }
    out = gzopen( gzfilename, "w" );
    if( out == NULL ) {
        Err( "Cannot Open %s.", gzfilename );
        fclose( in );
        return 1;
    }
    for (;;) {
        len = fread( buf, 1, sizeof(buf), in );
        if( ferror( in ) ) {
            Err( "fread." );
            return 1;
        }       
        if( len == 0 ) break;
        if( gzwrite( out, buf, (unsigned)len ) != len ) {
            Err( "gzwrite." );
            return 1;
        }
    }
    fclose( in );
    gzclose( out );
    return 0;
} /* GZip */


/* ** Func FileOut ***************************** */
/*
 * Open given file and, append given string to the file.
 * Check the filesize when it is oepned. If the file size
 * is big enough, rotate the file(rename and open new file to write).
 * return 0 --- error
 * return 1 --- success
 */
int FileOut( char *filename, char *string )
{
/* -- Local Data --------------------------------- */
    FILE *fp;
    long size;
/* -- Operation ---------------------------------- */
    fp = fopen( filename, "a" );
    if( fp == NULL ) { /* err */
        return 1;
    }
    fprintf( fp, "%s", string );
    size = ftell( fp );
    if( size < 0 ) { /* err */
        fclose( fp );
        return 1;
    }
    fclose( fp );
    if( size > LOGFILESIZE ) { /* rotate log file */
        int i;
        char log[512];
        char nextlog[512];
        /* remove oldest log file */
        sprintf( log, "%s.%d.gz", filename, LOGROTATENUM );
        remove( log );
        /* shift log files */
        for( i = LOGROTATENUM; i > 0; i-- ) {
            sprintf( nextlog, "%s.%d.gz", filename, i );
            sprintf( log, "%s.%d.gz", filename, i - 1 );
            rename( log, nextlog );
        }
        /* rename newest log file */
        sprintf( nextlog, "%s.%d", filename, 0 );
        sprintf( log, "%s", filename );
        rename( log, nextlog );
        GZip( nextlog );
    }
    return 0;
} /* FileOut */

