/* ** game.c ************************************* */
/* Functions which support Games. */
/* *********************************************** */

/* -- Include File ------------------------------- */
#include <config.h>
#include <stdlib.h>
#include "game.h"
#include "clientlist.h"
#include "handlist.h"
#include "sendmessage.h"
#include "tuitated.h"

/* -- Global Data -------------------------------- */
GAMEDATA *GameList;

/* -- Prototype Declerations  -------------------- */
void InitGameStructure( GAMEDATA *game );
int CleanViewer( GAMEDATA *game );
int DecideGameId( void );

/* ** Func InitGame ****************************** */
/*
 * Initialize for game.
 * Return 0 .. OK.
 * Return 1 .. Error.
 */
int InitGame( CLIENTDATA *player1, CLIENTDATA *player2 )
{
/* -- Local Data --------------------------------- */
    GAMEDATA *game;
    GAMEDATA *p;
/* -- Operation ---------------------------------- */
    /* allocate game data */
    game = malloc( sizeof( GAMEDATA ) );
    if( game == NULL ) {
        Err( "in InitGame. malloc." );
        return 1;
    }
    game->Black = player1;
    game->White = player2;
    InitGameStructure( game );

    /* save time this game is started */
    game->Begin = time( NULL );
    
    /* Send Game Start Message to Client */
    if( SendStart( player1, BLACK ) ) {
        DeleteClient( player1 );
        RemoveClient( player2 );
        return 1;
    }
    if( SendStart( player2, WHITE ) ) {
        RemoveClient( player1 );
        DeleteClient( player2 );
        return 1;
    }

    /* add to GameList */
    if( GameList == NULL ) {
        GameList = game;
        game->Next = NULL;
    } else {
        for( p = GameList; p->Next != NULL; p = p->Next );
        p->Next = game;
        game->Next = NULL;
    }

    /* set player mode */
    player1->mode = MODE_GAME;
    player1->game = game;
    player2->mode = MODE_GAME;
    player2->game = game;

    ServiceLog( "New Game %d - <Black %d %s> <White %d %s>",
                game->id,
                player1->soc, player1->name,
                player2->soc, player2->name );

    return 0;
} /* InitGame */

/* ** Func InitGameStructure ************************* */
/*
 * Initialize given game structure.
 */
void InitGameStructure( GAMEDATA *game )
{
/* -- Local Data --------------------------------- */
    int i, j;
    BOX *box;
/* -- Operation ---------------------------------- */
    game->id = DecideGameId();
    game->CurrentPlayer = game->Black;
    game->NonCurrentPlayer = game->White;
    game->ActionNumber = 0;

    for( i = 0; i < 9; i++ ) {
        for( j = 0; j < 9; j++ ) {
            box = &game->Board[i][j];
            box->Belong = NULL;
            box->PieceType = NONE;
            box->isPromoted = 0;
        }
    }

    game->Board[0][0].Belong = game->White;
    game->Board[0][0].PieceType = LANCE;
    game->Board[1][0].Belong = game->White;
    game->Board[1][0].PieceType = KNIGHT;
    game->Board[2][0].Belong = game->White;
    game->Board[2][0].PieceType = SILVER;
    game->Board[3][0].Belong = game->White;
    game->Board[3][0].PieceType = GOLD;
    game->Board[4][0].Belong = game->White;
    game->Board[4][0].PieceType = KING;
    game->Board[5][0].Belong = game->White;
    game->Board[5][0].PieceType = GOLD;
    game->Board[6][0].Belong = game->White;
    game->Board[6][0].PieceType = SILVER;
    game->Board[7][0].Belong = game->White;
    game->Board[7][0].PieceType = KNIGHT;
    game->Board[8][0].Belong = game->White;
    game->Board[8][0].PieceType = LANCE;
    game->Board[1][1].Belong = game->White;
    game->Board[1][1].PieceType = BISHOP;
    game->Board[7][1].Belong = game->White;
    game->Board[7][1].PieceType = ROOK;
    game->Board[0][2].Belong = game->White;
    game->Board[0][2].PieceType = PAWN;
    game->Board[1][2].Belong = game->White;
    game->Board[1][2].PieceType = PAWN;
    game->Board[2][2].Belong = game->White;
    game->Board[2][2].PieceType = PAWN;
    game->Board[3][2].Belong = game->White;
    game->Board[3][2].PieceType = PAWN;
    game->Board[4][2].Belong = game->White;
    game->Board[4][2].PieceType = PAWN;
    game->Board[5][2].Belong = game->White;
    game->Board[5][2].PieceType = PAWN;
    game->Board[6][2].Belong = game->White;
    game->Board[6][2].PieceType = PAWN;
    game->Board[7][2].Belong = game->White;
    game->Board[7][2].PieceType = PAWN;
    game->Board[8][2].Belong = game->White;
    game->Board[8][2].PieceType = PAWN;

    game->Board[0][8].Belong = game->Black;
    game->Board[0][8].PieceType = LANCE;
    game->Board[1][8].Belong = game->Black;
    game->Board[1][8].PieceType = KNIGHT;
    game->Board[2][8].Belong = game->Black;
    game->Board[2][8].PieceType = SILVER;
    game->Board[3][8].Belong = game->Black;
    game->Board[3][8].PieceType = GOLD;
    game->Board[4][8].Belong = game->Black;
    game->Board[4][8].PieceType = KING;
    game->Board[5][8].Belong = game->Black;
    game->Board[5][8].PieceType = GOLD;
    game->Board[6][8].Belong = game->Black;
    game->Board[6][8].PieceType = SILVER;
    game->Board[7][8].Belong = game->Black;
    game->Board[7][8].PieceType = KNIGHT;
    game->Board[8][8].Belong = game->Black;
    game->Board[8][8].PieceType = LANCE;
    game->Board[7][7].Belong = game->Black;
    game->Board[7][7].PieceType = BISHOP;
    game->Board[1][7].Belong = game->Black;
    game->Board[1][7].PieceType = ROOK;
    game->Board[0][6].Belong = game->Black;
    game->Board[0][6].PieceType = PAWN;
    game->Board[1][6].Belong = game->Black;
    game->Board[1][6].PieceType = PAWN;
    game->Board[2][6].Belong = game->Black;
    game->Board[2][6].PieceType = PAWN;
    game->Board[3][6].Belong = game->Black;
    game->Board[3][6].PieceType = PAWN;
    game->Board[4][6].Belong = game->Black;
    game->Board[4][6].PieceType = PAWN;
    game->Board[5][6].Belong = game->Black;
    game->Board[5][6].PieceType = PAWN;
    game->Board[6][6].Belong = game->Black;
    game->Board[6][6].PieceType = PAWN;
    game->Board[7][6].Belong = game->Black;
    game->Board[7][6].PieceType = PAWN;
    game->Board[8][6].Belong = game->Black;
    game->Board[8][6].PieceType = PAWN;

    game->BlackPiecesInHand.Pawn = 0;
    game->BlackPiecesInHand.Lance = 0;
    game->BlackPiecesInHand.Knight = 0;
    game->BlackPiecesInHand.Silver = 0;
    game->BlackPiecesInHand.Gold = 0;
    game->BlackPiecesInHand.Bishop = 0;
    game->BlackPiecesInHand.Rook = 0;

    game->WhitePiecesInHand.Pawn = 0;
    game->WhitePiecesInHand.Lance = 0;
    game->WhitePiecesInHand.Knight = 0;
    game->WhitePiecesInHand.Silver = 0;
    game->WhitePiecesInHand.Gold = 0;
    game->WhitePiecesInHand.Bishop = 0;
    game->WhitePiecesInHand.Rook = 0;

    game->BlackLifeNum = 8;
    game->WhiteLifeNum = 8;

    game->HandList = NULL;
    game->ViewerList = NULL;

} /* InitGameStructure */

/* ** Func CleanGame ****************************** */
/*
 * Cleanup structures for a game.
 * Cleanup figters and viewers data structures too,
 * and close connections with them.
 * Return 0 .. OK.
 * Return 1 .. Error.
 */
int CleanGame( GAMEDATA *game )
{
/* -- Local Data --------------------------------- */
    GAMEDATA *p;
/* -- Operation ---------------------------------- */
    ServiceLog( "Game End %d", game->id );
    
    /*  from GameList */
    if( GameList == game ) {
        GameList = game->Next;
    } else {
        p = GameList;
        while( 1 ) {
            if( p->Next == NULL ) return 1;
            if( p->Next == game ) break;
            p = p->Next;
        }
        p->Next = p->Next->Next;
    }

    RemoveClient( game->Black );
    RemoveClient( game->White );
    CleanHand( game );
    CleanViewer( game );

    free( game );

    return 0;
} /* CleanGame */

/* ** Func CleanGameWithEmergence ************************ */
/*
 * Cleanup structures for a game.
 * Cleanup figters and viewers data structures too,
 * and close connections with them.
 * Given client means that the player should be forced to be removed.
 * Return 0 .. OK.
 * Return 1 .. Error.
 */
int CleanGameWithEmergence( GAMEDATA *game, CLIENTDATA *client )
{
/* -- Local Data --------------------------------- */
    GAMEDATA *p;
/* -- Operation ---------------------------------- */
    ServiceLog( "Game End %d", game->id );
    
    /*  from GameList */
    if( GameList == game ) {
        GameList = game->Next;
    } else {
        p = GameList;
        while( 1 ) {
            if( p->Next == NULL ) return 1;
            if( p->Next == game ) break;
            p = p->Next;
        }
        p->Next = p->Next->Next;
    }

    if( client == game->Black ) {
        DeleteClient( game->Black );
        RemoveClient( game->White );
    } else {
        RemoveClient( game->Black );
        DeleteClient( game->White );
    }
    CleanHand( game );
    CleanViewer( game );

    free( game );

    return 0;
} /* CleanGameWithEmergence */

/* ** Func FindClientSelectingEnemy ************** */
/*
 * Find Client which is selecting enemy in list,
 * and return it. If not found, return NULL.
 */
CLIENTDATA *FindClientSelectingEnemy( CLIENTDATA *me )
{
/* -- Local Data --------------------------------- */
    CLIENTDATA *p;
/* -- Operation ---------------------------------- */
    p = ClientList;
    while( p != NULL ) {
        if( p->mode == MODE_SELECTENEMY && p != me ) {
            return p;
        }
        p = p->Next;
    }
    return NULL;
} /* FindClientSelectingEnemy */

/* ** Func FlipPlayer ************************ */
/*
 * Flip Current Playing Player.
 */
void FlipPlayer( GAMEDATA *game )
{
/* -- Local Data --------------------------------- */
    CLIENTDATA *p;
/* -- Operation ---------------------------------- */
    p = game->CurrentPlayer;
    game->CurrentPlayer = game->NonCurrentPlayer;
    game->NonCurrentPlayer = p;
} /* FlipPlayer */

/* ** Func AddViewer ***************************** */
/*
 * Add given client to Viewer List of the game.
 */
void AddViewer( GAMEDATA *game, CLIENTDATA *ViewClient )
{
/* -- Local Data --------------------------------- */
    VIEWERDATA *p;
    VIEWERDATA *viewer;
/* -- Operation ---------------------------------- */
    viewer = malloc( sizeof( VIEWERDATA ) );
    viewer->Next = NULL;
    viewer->client = ViewClient;
    ViewClient->game = game;
    if( game->ViewerList == NULL ) {
        game->ViewerList = viewer;
    } else {
        p = game->ViewerList;
        while( p->Next != NULL ) p = p->Next;
        p->Next = viewer;
    }
    ServiceLog( "New Viewer %d ( Game %d )", ViewClient->soc, game->id );
} /* AddViewer */

/* ** Func CleanViewer ***************************** */
/*
 * Remove all clients of Viewer List of the game.
 */
int CleanViewer( GAMEDATA *game )
{
/* -- Local Data --------------------------------- */
    VIEWERDATA *p;
    VIEWERDATA *next;
/* -- Operation ---------------------------------- */
    p = game->ViewerList;
    while( p != NULL ) {
        next = p->Next;
        RemoveClient( p->client );
        free( p );
        p = next;
    }
    return 0;
} /* CleanViewer */

/* ** Func RemoveViewer ***************************** */
/*
 * Remove given client from List of the game.
 */
int RemoveViewer( GAMEDATA *game, CLIENTDATA *client )
{
/* -- Local Data --------------------------------- */
    VIEWERDATA *p;
    VIEWERDATA *next;
/* -- Operation ---------------------------------- */
    ServiceLog( "Viewer %d end view ( game %d )", client->soc, game->id );
    if( game->ViewerList->client == client ) {
        next = game->ViewerList->Next;
        RemoveClient( game->ViewerList->client );
        free( game->ViewerList );
        game->ViewerList = next;
    } else {
        p = game->ViewerList;
        next = p->Next;
        while( next != NULL ) {
            if( next->client == client ) {
                p->Next = next->Next;
                RemoveClient( next->client );
                free( next );
                return 0;
            }
            p = next;
            next = p->Next;
        }
    }
    return 1;
} /* RemoveViewer */


/* ** Func DeleteViewer ***************************** */
/*
 * Force given client to be removed from List of the game.
 */
int DeleteViewer( GAMEDATA *game, CLIENTDATA *client )
{
/* -- Local Data --------------------------------- */
    VIEWERDATA *p;
    VIEWERDATA *next;
/* -- Operation ---------------------------------- */
    ServiceLog( "Viewer %d end view ( game %d )", client->soc, game->id );
    if( game->ViewerList->client == client ) {
        next = game->ViewerList->Next;
        DeleteClient( game->ViewerList->client );
        free( game->ViewerList );
        game->ViewerList = next;
    } else {
        p = game->ViewerList;
        next = p->Next;
        while( next != NULL ) {
            if( next->client == client ) {
                p->Next = next->Next;
                DeleteClient( next->client );
                free( next );
                return 0;
            }
            p = next;
            next = p->Next;
        }
    }
    return 1;
} /* DeleteViewer */

/* ** Func GameNum ***************************** */
/*
 * return number of Current Games.
 */
int GameNum( void )
{
/* -- Local Data --------------------------------- */
    int ret = 0;
    GAMEDATA *p;
/* -- Operation ---------------------------------- */
    p = GameList;
    while( p != NULL ) {
        p = p->Next;
        ret++;
    }
    return ret;
} /* GameNum */

/* ** Func MatchGameId *************************** */
/*
 * return number of Current Games.
 */
GAMEDATA *MatchGameId( int id )
{
/* -- Local Data --------------------------------- */
    GAMEDATA *p;
/* -- Operation ---------------------------------- */
    p = GameList;
    while( p != NULL ) {
        if( p->id == id ) return p;
        p = p->Next;
    }
    return NULL;
} /* MatchGameId */

/* ** Func DecideGameId *************************** */
/*
 * return number of Game id which is not used now.
 */
int DecideGameId( void )
{
/* -- Local Data --------------------------------- */
    int ret = 1;
    int isDecided = FALSE;
    int found;
    GAMEDATA *p;
/* -- Operation ---------------------------------- */
    while( !isDecided ) {
        found = FALSE;
        p = GameList;
        while( p != NULL ) {
            if( p->id == ret ) {
                found = TRUE;
                break;
            }
            p = p->Next;
        }
        if( !found ) {
            isDecided = TRUE;
        } else {
            ret++;
        }
    }
    return ret;
} /* DecideGameId */




