/* ** handlist.c ********************************* */
/* Routines which handle hand data. */
/* *********************************************** */

/* -- Include File ------------------------------- */
#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include "protocol.h"
#include "struct.h"
#include "notation.h"
#include "handlist.h"
#include "sendmessage.h"
#include "tuitated.h"

/* -- Global Data -------------------------------- */
const char StrBLACK[] = "Black";
const char StrWHITE[] = "White";

extern char NotationFile[512];

/* -- Protype Declerations  ---------------------- */
extern int ManageNotation( void );
extern int GenerateNotationID( void );

/* ** Func AddHand ***************************** */
/*
 * Add given Hand to given game structure's handlist.
 * return 0... Success.
 * return 1... Fail.
 */
int AddHand( GAMEDATA *game, const NOTATION_HANDDATA *data )
{
/* -- Local Data --------------------------------- */
    NOTATION_HANDDATA *p;
/* -- Operation ---------------------------------- */
    if( game->HandList == NULL ) {
        game->HandList =(NOTATION_HANDDATA *)malloc(sizeof(NOTATION_HANDDATA));
        if( game->HandList == NULL ) {
            Err( "malloc." );
            return 1;
        }
        /* Err( "Adding FIRST Hand.. %x", game->HandList ); */
        memcpy( game->HandList, data, sizeof( NOTATION_HANDDATA ) );
        game->HandList->Next = NULL;
    } else {
        p = game->HandList;
        while( p->Next != NULL ) {
            p = p->Next;
        }
        p->Next = (NOTATION_HANDDATA *)malloc( sizeof( NOTATION_HANDDATA ) );
        if( p->Next == NULL ) {
            Err( "malloc." );
            return 1;
        }
        /* Err( "Adding Hand.. %x", p->Next ); */
        memcpy( p->Next, data, sizeof( NOTATION_HANDDATA ) );
        p->Next->Next = NULL;
    }
    return 0;
} /* AddHand */

/* ** Func CleanHand ***************************** */
/*
 * Cleanup List of Hand of given game structure.
 * return 0... Success.
 * return 1... Fail.
 */
int CleanHand( const GAMEDATA *game )
{
/* -- Local Data --------------------------------- */
    NOTATION_HANDDATA *p;
    NOTATION_HANDDATA *next;
/* -- Operation ---------------------------------- */
    p = game->HandList;
    while( p != NULL ) {
        next = p->Next;
        free( p );
        p = next;
    }
    return 0;
} /* CleanHand */


/* ** Func SaveNotation ************************ */
/*
 * SaveNotation of given game.
 */
int SaveNotation( const GAMEDATA *game )
{
/* -- Local Data --------------------------------- */
    char time_str[256];
    char buf[256];
    FILE *fp;
    time_t Time;
    NOTATION_HANDDATA *hand;
    int id;
    NOTATIONDATA *notation;
    long pos;
/* -- Operation ---------------------------------- */
    ManageNotation();
    Time = time( NULL );
    strcpy( time_str, ctime( &Time ) );
    id = GenerateNotationID();
    if( id == 0 ) {
        Err( "Fail to generate notation ID." );
        return 1;
    }
    /* write into notation file */
    fp = fopen( NotationFile, "a" );
    if( fp == NULL ) {
        Err( "CannotOpen NotationFile -- %s.", NotationFile );
        return 1;
    }
    /* Magic */
    fprintf( fp, MAGIC );
    /* ID */
    fprintf( fp, " %d\n", id );
    /* Begin time and End time */
    strcpy( time_str, ctime( &game->Begin ) );
    fprintf( fp, "%ld %s", game->Begin, time_str );
    strcpy( time_str, ctime( &Time ) );
    fprintf( fp, "%ld %s", Time, time_str );
    /* Name and MailAddress */
    fprintf( fp, "%s %s %s %s\n",
             game->Black->name,
             game->Black->mailaddress,
             game->White->name,
             game->White->mailaddress );
    /* save file position to reread */
    pos = ftell( fp );
    /* write hand list */
    hand = game->HandList;
    while( hand != NULL ) {
        HandDataToMsg( hand, buf );
        fprintf( fp, "%s", buf );
        hand = hand->Next;
    }
    fclose( fp );
    /* add notation list */
    notation = malloc( sizeof( NOTATIONDATA ) );
    if( notation == NULL ) {
        Err( "malloc." );
        return 1;
    }
    notation->Next = NULL;
    notation->ID = id;
    notation->saved = FALSE;
    notation->Begin = game->Begin;
    notation->End = Time;
    /* name and address */
    notation->BlackName = malloc( strlen( game->Black->name ) + 1 );
    if( notation->BlackName == NULL ) {
        Err( "malloc.");
        free( notation );
        return 1;
    }
    strcpy( notation->BlackName, game->Black->name );
    
    notation->BlackAddress = malloc( strlen( game->Black->mailaddress ) + 1 );
    if( notation->BlackAddress == NULL ) {
        Err( "malloc.");
        free( notation->BlackName );
        free( notation );
        return 1;
    }
    strcpy( notation->BlackAddress, game->Black->mailaddress );
    
    notation->WhiteName = malloc( strlen( game->White->name ) + 1 );
    if( notation->WhiteName == NULL ) {
        Err( "malloc.");
        free( notation->BlackName );
        free( notation->BlackAddress );
        free( notation );
        return 1;
    }
    strcpy( notation->WhiteName, game->White->name );
    
    notation->WhiteAddress = malloc( strlen( game->White->mailaddress ) + 1 );
    if( notation->WhiteAddress == NULL ) {
        Err( "malloc.");
        free( notation->BlackName );
        free( notation->BlackAddress );
        free( notation->WhiteName );
        free( notation );
        return 1;
    }
    strcpy( notation->WhiteAddress, game->White->mailaddress );
    /* file position */
    notation->NotationPos = pos;
    /* link */
    if( NotationList == NULL ) {
        NotationList = notation;
    } else {
        NOTATIONDATA *cur;
        cur = NotationList;
        while( cur->Next != NULL )
            cur = cur->Next;
        cur->Next = notation;
    }
    return 0;
} /* SaveNotation */


/* ** Func HandNum ******************************* */
/*
 * Return current Number of Hand of given game.
 */
int HandNum( const GAMEDATA *game )
{
/* -- Local Data --------------------------------- */
    int ret = 0;
    NOTATION_HANDDATA *p;
/* -- Operation ---------------------------------- */
    p = game->HandList;
    while( p != NULL ) {
        p = p->Next;
        ret++;
    }
    return ret;
} /* HandNum */




