/* ** init.c ************************************* */
/* Initialize tuitated. */
/* *********************************************** */

/* -- Include File ------------------------------- */
#include <config.h>
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/signal.h>
#include <sys/param.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include "notation.h"
#include "receivemessage.h"
#include "tuitated.h"

/* -- Global Data -------------------------------- */
extern short port;
extern int InDebugMode;
extern int ApplySocketAlreadyCreated;
extern int ShouldReopenApplySocket;


/* -- Protype Declerations  ---------------------- */
void Init( void );
void InitSigHup( void );
void InitSigPipe( void );
void InitSigChild( void );
void InitSigTerm( void );
void InitDaemon( void );
int InitSoc( void );
int TestSoc( void );

extern int ParseConfigFile( void ); /* in main.c */
extern void TerminateAllGames( void ); /* in apply.c */

/* ** Func Init ********************************** */
/* Initialize tuitated. */
void Init( void )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
    InitNotationList();
    if( !InDebugMode )	InitDaemon();
    if( InitSoc() ) {
        Err( "Socket initialize failure." );
        exit( 1 );
    }
    InitReceiveHandlers();
    InitSigHup();
    InitSigPipe();
    InitSigChild();
    InitSigTerm();
    UpdateNotationList();
} /* Init */

/* ** Func SigHup ******************************** */
/*
 * Signal Handler for SIGHUP.
 * Relead configuration file.
 */
void SigHup( int dummy )
{
    ParseConfigFile();
    if( ShouldReopenApplySocket ) {
        close( ApplySocket );
        if( InitSoc() ) {
            Err( "Socket reinitialize for SIGHUP failure." );
            exit( 1 );
        }
        ShouldReopenApplySocket = FALSE;
    }
    UpdateNotationList();
    InitSigHup(); /* Reset SIGHUP handler for some Systems. */
} /* SigHup */

/* ** Func SigPipe ******************************** */
/*
 * Signal Handler for SIGHUP.
 * Relead configuration file.
 */
void SigPipe( int dummy )
{
    InitSigPipe(); /* Reset SIGPIPE handler for some Systems. */
    Err( "SIGPIPE received." );
} /* SigPipe */

/* ** Func SigChild ****************************** */
/* Handler for Signal "SIGCHLD". */
void SigChild( int dummy )
{
/* -- Local Data --------------------------------- */
    int pid;
    int status;
/* -- Operation ---------------------------------- */
    signal( SIGCHLD, SIG_IGN );

    while( ( pid = wait3( &status, WNOHANG, (struct rusage *) 0 ) ) > 0 )
        ;

    signal( SIGCHLD, SigChild );

    return;
} /* SigChild */

/* ** Func SigTerm ****************************** */
/* Handler for Signal "SIGTERM". */
void SigTerm( int dummy )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
    Err( "SIGTERM received." );
    TerminateAllGames();
    close( ApplySocket ); 
    exit( 0 );
} /* SigTerm */

/* ** Func InitSigHup **************************** */
/* Set Signal handler for SIGHUP  */
void InitSigHup( void )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
#ifdef SIGHUP
    if( signal( SIGHUP, SigHup ) < 0 ) {
        Err( "signal. (SIGHUP)" );
    }
#endif
} /* InitSigHup */

/* ** Func InitSigPipe **************************** */
/* Set Signal handler for SIGPIPE  */
void InitSigPipe( void )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
#ifdef SIGPIPE
    if( signal( SIGPIPE, SigPipe ) < 0 ) {
        Err( "signal. (SIGPIPE)" );
    }
#endif
} /* InitSigPipe */

/* ** Func InitSigChild **************************** */
/* Set Signal handler for SIGCHLD  */
void InitSigChild( void )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
#ifdef SIGCHLD
    if( signal( SIGCHLD, SigChild ) < 0 ) {
        Err( "signal. (SIGCHLD)" );
    }
#else /* !SIGCHLD */
#ifdef SIGCLD
    if( signal( SIGCLD, SigChild ) < 0 ) {
        Err( "signal. (SIGCLD)" );
    }
#endif  /* SIGCLD */
#endif  /* SIGCHLD */
} /* InitSigChild */


/* ** Func InitSigTerm **************************** */
/* Set Signal handler for SIGTERM  */
void InitSigTerm( void )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
#ifdef SIGTERM
    if( signal( SIGTERM, SigTerm ) < 0 ) {
        Err( "signal. (SIGTERM)" );
    }

#endif  /* SIGTERM */
} /* InitSigTerm */


/* ** Func InitDaemon **************************** */
/* Initialize Tuitated as a Daemon. */
void InitDaemon( void )
{
/* -- Local Data --------------------------------- */
    int ChildPid;
    int fd;
/* -- Operation ---------------------------------- */
    if( getppid() != 1 ) { /* if not executed by "init" process.... */
	/* Ignore some Signals */
#ifdef SIGTTOU
        signal( SIGTTOU, SIG_IGN );
#endif
#ifdef SIGTTIN
        signal( SIGTTIN, SIG_IGN );
#endif
#ifdef SIGTSTP
        signal( SIGTSTP, SIG_IGN );
#endif
        /* fork not to be Group Leader */
        if( (ChildPid = fork()) < 0 ) { /* Error */
            Err( "fork." );
        } else if( ChildPid > 0 ) {	/* Parent */
            exit( 0 );
        }	                        /* Child */
#ifdef SETPGRP_VOID
        if( setpgrp() == -1 ) {
            Err( "setpgrp." );
        }
#else
        if( setpgrp( 0, getpid() ) == -1 ) {
            Err( "setpgrp." );
        }
#endif
#ifdef TIOCNOTTY
        if( (fd = open( "/dev/tty", O_RDWR )) >= 0 ) {
            ioctl( fd, TIOCNOTTY, (char *)NULL );
            close( fd );
        }
#endif
    }

    /* close all fd */
    for( fd = 0; fd < NOFILE; fd++ )
        close( fd );

    errno = 0;
    chdir( "/" );
    umask( 0 );

} /* InitDaemon */


/* ** Func InitSoc ******************************* */
/*
 * Initialize the Socket for Tuitated.
 * The port number to use in global variable 'port'.
 * Opened socket fd is stored into
 * the global variable 'ApplySocket'.
 * Return 0 --- success.
 * Return 1 --- failure.
 * Do not exit on failure.
 */
int InitSoc( void )
{
/* -- Local Data --------------------------------- */
    struct hostent *serverhost;
    struct sockaddr_in server;
    char hostname[32];
    int optval;
    int iRet;
/* -- Operation ---------------------------------- */
    if( TestSoc() )
        return 1;
    if( gethostname( hostname, 32 ) == -1 ) {
        Err( "gethostname.\n" );
        return 1;
    }
    if( ( serverhost = gethostbyname( hostname ) ) == NULL ) {
        Err( "gethostbyname.\n" );
        return 1;
    }
    bzero( (char *)&server, sizeof(server) );
    server.sin_family = AF_INET;
    server.sin_port = htons( (port == 0) ? PORT : port );
    server.sin_addr.s_addr = htonl( INADDR_ANY );
    ApplySocket
        = socket( AF_INET, SOCK_STREAM, getprotobyname("tcp")->p_proto );
    if( ApplySocket < 0 ) {
        Err( "socket.\n" );
        return 1;
    }	
    optval = 1;
#ifdef SO_REUSEPORT
    if( InDebugMode ) {
        iRet = setsockopt( ApplySocket, SOL_SOCKET, SO_REUSEPORT,
                           &optval, sizeof( optval ) );
        if( iRet < 0 ) {
            Err( "setsockopt.\n" );
            return 1;
        }
    }
#endif
    iRet = bind( ApplySocket, (struct sockaddr *)&server, sizeof(server) );
    if( iRet == -1 ) {
        Err( "bind.\n" );
        return 1;
    }
    if( listen( ApplySocket, 5 ) == -1 ) {
        Err( "listen.\n" );
        return 1;
    }
    ApplySocketAlreadyCreated = TRUE;
    return 0;
} /* InitSoc */

/* ** Func TestSoc ******************************* */
/*
 * Test tuitated is already up or not.
 *  by tring to connect to the socket of tutiated.
 * Return 0 --- success.
 * Return 1 --- failure.
 * Do not exit on failure.
 */
int TestSoc( void )
{
/* -- Local Data --------------------------------- */
    int TestSoc;
    struct sockaddr_in server;
/* -- Operation ---------------------------------- */
    TestSoc = socket( AF_INET, SOCK_STREAM, getprotobyname("tcp")->p_proto );
    if( TestSoc < 0 ) {
        Err( "socket.\n" );
        return 1;
    }
    server.sin_family = AF_INET;
    server.sin_port = htons( (port == 0) ? PORT : port );
    server.sin_addr.s_addr = htonl( INADDR_ANY );
    if(connect( TestSoc, (struct sockaddr *)&server, sizeof( server ) ) < 0 ) {
        close( TestSoc );
    } else {
        close( TestSoc );
        Err( "The port %d is already used.", port );
        return 1;
    }
    return 0;
} /* TestSoc */

