/* ** judge.c ************************************ */
/* Judge each hand, and have current data. */
/* *********************************************** */

/* -- Include File ------------------------------- */
#include "judge.h"
#include "protocol.h"
#include "game.h"
#include "tuitated.h"

/* -- Global Data -------------------------------- */
/* -- Prototype Declerations  -------------------- */

/* ** Func ExamineHand *************************** */
/*
 * Examine Each hand for given game,
 * and return JUDGERESULT.
 * If it is needed, change current data and
 * save historical data, or do clean up game data
 * to terminate game if it is needed.
 * MOVE and DROP is handled with this func.
 */


#define NASHI ( CLIENTDATA * ) 0
#define JIBUN ( CLIENTDATA * ) 1
#define TEKI ( CLIENTDATA * ) 2

void Cheat( GAMEDATA *game, int x, int y, PIECE *pPiece, SIDE *pSide,
		   int *pIsPromoted )
{
	*pPiece	= game->Board[x-1][y-1].PieceType;


	if( *pPiece == NONE ){
		*pSide = NOSIDE;
	}
	else{
		if( game->Board[x-1][y-1].Belong == game->Black )*pSide = BLACK;
		if( game->Board[x-1][y-1].Belong == game->White	)*pSide = WHITE;
	}

	*pIsPromoted	= game->Board[x-1][y-1].isPromoted;

	return;
}


JUDGERESULT ExamineHand
(
	GAMEDATA *game,
	const CLIENT_MESSAGE *data,
	JUDGE *judge
)
{
	JUDGERESULT kekka;

	kekka = ExamineHandSub( game, data , judge );

	if( kekka == VIOLATE ){
		if( game->CurrentPlayer == game->Black ){
			game->BlackLifeNum --;
			if( game->BlackLifeNum < 0 ){
				judge->result = END;
				judge->endstate.winner = game->NonCurrentPlayer;
				judge->endstate.reason = LOSTLIFE;
				return END;
			}
		}
		if( game->CurrentPlayer == game->White ){
			game->WhiteLifeNum --;
			if( game->WhiteLifeNum < 0 ){
				judge->result = END;
				judge->endstate.winner = game->NonCurrentPlayer;
				judge->endstate.reason = LOSTLIFE;
				return END;
			}
		}
	}
	if( kekka == OK ){
		switch( data->key ){

			case MOVE:
				judge->isCaptured = 1;
				judge->capture.piece = game->Board[data->hand.move.dx]
				[data->hand.move.dy].PieceType;
				judge->capture.x = data->hand.move.dx;
				judge->capture.y = data->hand.move.dy;
				judge->capture.hasPromoted = game->Board[data->hand.move.dx]
				[data->hand.move.dy].isPromoted;

				switch( game->Board[data->hand.move.dx][data->hand.move.dy]
					   .PieceType ){

					case PAWN:
						if( game->CurrentPlayer == game->Black ){
							game->BlackPiecesInHand.Pawn++;
						}
						if( game->CurrentPlayer == game->White ){
							game->WhitePiecesInHand.Pawn++;
						}
						break;

					case LANCE:
						if( game->CurrentPlayer == game->Black ){
							game->BlackPiecesInHand.Lance++;
						}
						if( game->CurrentPlayer == game->White ){
							game->WhitePiecesInHand.Lance++;
						}
						break;

					case KNIGHT:
						if( game->CurrentPlayer == game->Black ){
							game->BlackPiecesInHand.Knight++;
						}
						if( game->CurrentPlayer == game->White ){
							game->WhitePiecesInHand.Knight++;
						}
						break;

					case SILVER:
						if( game->CurrentPlayer == game->Black ){
							game->BlackPiecesInHand.Silver++;
						}
						if( game->CurrentPlayer == game->White ){
							game->WhitePiecesInHand.Silver++;
						}
						break;

					case GOLD:
						if( game->CurrentPlayer == game->Black ){
							game->BlackPiecesInHand.Gold++;
						}
						if( game->CurrentPlayer == game->White ){
							game->WhitePiecesInHand.Gold++;
						}
						break;

					case BISHOP:
						if( game->CurrentPlayer == game->Black ){
							game->BlackPiecesInHand.Bishop++;
						}
						if( game->CurrentPlayer == game->White ){
							game->WhitePiecesInHand.Bishop++;
						}
						break;

					case ROOK:
						if( game->CurrentPlayer == game->Black ){
							game->BlackPiecesInHand.Rook++;
						}
						if( game->CurrentPlayer == game->White ){
							game->WhitePiecesInHand.Rook++;
						}
						break;

					default:
					judge->isCaptured = 0;
						break;
				}
				game->Board[data->hand.move.dx][data->hand.move.dy].Belong
				= game->Board[data->hand.move.sx][data->hand.move.sy].Belong;
				game->Board[data->hand.move.dx][data->hand.move.dy].PieceType
				= game->Board[data->hand.move.sx][data->hand.move.sy]
				.PieceType;
				game->Board[data->hand.move.dx][data->hand.move.dy].isPromoted
				= game->Board[data->hand.move.sx][data->hand.move.sy]
				.isPromoted;
				if( data->hand.move.willPromote ){
					game->Board[data->hand.move.dx][data->hand.move.dy]
					.isPromoted = 1;
				}
				game->Board[data->hand.move.sx][data->hand.move.sy].Belong
				= NULL;
				game->Board[data->hand.move.sx][data->hand.move.sy].PieceType
				= NONE;
				break;	

			case DROP:
				switch( data->hand.drop.piece ){

					case PAWN:
						if( game->CurrentPlayer == game->Black ){
							if(game->BlackPiecesInHand.Pawn){
								game->BlackPiecesInHand.Pawn--;
							}
							else{
								return ILLEGAL;
							}
						}
						if( game->CurrentPlayer == game->White ){
							if(game->WhitePiecesInHand.Pawn){
								game->WhitePiecesInHand.Pawn--;
							}
							else{
								return ILLEGAL;
							}
						}
						break;

					case LANCE:
						if( game->CurrentPlayer == game->Black ){
							if(game->BlackPiecesInHand.Lance){
								game->BlackPiecesInHand.Lance--;
							}
							else{
								return ILLEGAL;
							}
						}
						if( game->CurrentPlayer == game->White ){
							if(game->WhitePiecesInHand.Lance){
								game->WhitePiecesInHand.Lance--;
							}
							else{
								return ILLEGAL;
							}
						}
						break;

					case KNIGHT:
						if( game->CurrentPlayer == game->Black ){
							if(game->BlackPiecesInHand.Knight){
								game->BlackPiecesInHand.Knight--;
							}
							else{
								return ILLEGAL;
							}
						}
						if( game->CurrentPlayer == game->White ){
							if(game->WhitePiecesInHand.Knight){
								game->WhitePiecesInHand.Knight--;
							}
							else{
								return ILLEGAL;
							}
						}
						break;

					case SILVER:
						if( game->CurrentPlayer == game->Black ){
							if(game->BlackPiecesInHand.Silver){
								game->BlackPiecesInHand.Silver--;
							}
							else{
								return ILLEGAL;
							}
						}
						if( game->CurrentPlayer == game->White ){
							if(game->WhitePiecesInHand.Silver){
								game->WhitePiecesInHand.Silver--;
							}
							else{
								return ILLEGAL;
							}
						}
						break;

					case GOLD:
						if( game->CurrentPlayer == game->Black ){
							if(game->BlackPiecesInHand.Gold){
								game->BlackPiecesInHand.Gold--;
							}
							else{
								return ILLEGAL;
							}
						}
						if( game->CurrentPlayer == game->White ){
							if(game->WhitePiecesInHand.Gold){
								game->WhitePiecesInHand.Gold--;
							}
							else{
								return ILLEGAL;
							}
						}
						break;

					case BISHOP:
						if( game->CurrentPlayer == game->Black ){
							if(game->BlackPiecesInHand.Bishop){
								game->BlackPiecesInHand.Bishop--;
							}
							else{
								return ILLEGAL;
							}
						}
						if( game->CurrentPlayer == game->White ){
							if(game->WhitePiecesInHand.Bishop){
								game->WhitePiecesInHand.Bishop--;
							}
							else{
								return ILLEGAL;
							}
						}
						break;

					case ROOK:
						if( game->CurrentPlayer == game->Black ){
							if(game->BlackPiecesInHand.Rook){
								game->BlackPiecesInHand.Rook--;
							}
							else{
								return ILLEGAL;
							}
						}
						if( game->CurrentPlayer == game->White ){
							if(game->WhitePiecesInHand.Rook){
								game->WhitePiecesInHand.Rook--;
							}
							else{
								return ILLEGAL;
							}
						}
						break;

					default:
						break;
				}
				game->Board[data->hand.drop.dx][data->hand.drop.dy].Belong
				= game->CurrentPlayer;
				game->Board[data->hand.drop.dx][data->hand.drop.dy].PieceType
				= data->hand.drop.piece;
				game->Board[data->hand.drop.dx][data->hand.drop.dy].isPromoted
				= 0;
				break;

			default:
				return ILLEGAL;
		}
	}

	judge->result = kekka;
	return kekka;

}/* ExamineHand */


JUDGERESULT ExamineHandSub
(
	GAMEDATA *game,
	const CLIENT_MESSAGE *data,
	JUDGE *judge
)
{

	int i,j;
	GAMEDATA Sight_game;
	CLIENT_MESSAGE Sight_data;

	JUDGERESULT flag;

/* judge ν */
	judge->isCheck = 0;
	judge->isCaptured = 0;
	judge->capture.piece = NONE;
	
	if( game->CurrentPlayer == game->Black ){
		for( i = 0 ; i < 9 ; i ++ ){
			for( j = 0 ; j < 9 ; j ++ ){
				if( game->Board[i][j].Belong == game->CurrentPlayer ){
					Sight_game.Board[i][j].Belong = JIBUN;
				}
				else if( game->Board[i][j].Belong == game->NonCurrentPlayer ){
					Sight_game.Board[i][j].Belong = TEKI;
				}
				else {
					Sight_game.Board[i][j].Belong = NASHI;
				}
				Sight_game.Board[i][j].PieceType = game->Board[i][j].PieceType;
				Sight_game.Board[i][j].isPromoted
				= game->Board[i][j].isPromoted;
			}
		}
		Sight_game.BlackPiecesInHand.Pawn = game->BlackPiecesInHand.Pawn;
		Sight_game.BlackPiecesInHand.Lance = game->BlackPiecesInHand.Lance;
		Sight_game.BlackPiecesInHand.Knight = game->BlackPiecesInHand.Knight;
		Sight_game.BlackPiecesInHand.Silver = game->BlackPiecesInHand.Silver;
		Sight_game.BlackPiecesInHand.Gold = game->BlackPiecesInHand.Gold;
		Sight_game.BlackPiecesInHand.Bishop = game->BlackPiecesInHand.Bishop;
		Sight_game.BlackPiecesInHand.Rook = game->BlackPiecesInHand.Rook;

		Sight_game.WhitePiecesInHand.Pawn = game->WhitePiecesInHand.Pawn;
		Sight_game.WhitePiecesInHand.Lance = game->WhitePiecesInHand.Lance;
		Sight_game.WhitePiecesInHand.Knight = game->WhitePiecesInHand.Knight;
		Sight_game.WhitePiecesInHand.Silver = game->WhitePiecesInHand.Silver;
		Sight_game.WhitePiecesInHand.Gold = game->WhitePiecesInHand.Gold;
		Sight_game.WhitePiecesInHand.Bishop = game->WhitePiecesInHand.Bishop;
		Sight_game.WhitePiecesInHand.Rook = game->WhitePiecesInHand.Rook;

		Sight_data.key = data->key;
		switch( data->key ){

			case MOVE:
				Sight_data.hand.move.sx = data->hand.move.sx;
				Sight_data.hand.move.sy = data->hand.move.sy;
				Sight_data.hand.move.dx = data->hand.move.dx;
				Sight_data.hand.move.dy = data->hand.move.dy;
				Sight_data.hand.move.willPromote
				= data->hand.move.willPromote;
				break;

			case DROP:
				Sight_data.hand.drop.dx = data->hand.drop.dx;
				Sight_data.hand.drop.dy = data->hand.drop.dy;
				Sight_data.hand.drop.piece = data->hand.drop.piece;
				break;

			default:
				break;
		}
	}else if( game->CurrentPlayer == game->White ){
		for( i = 0 ; i < 9 ; i ++ ){
			for( j = 0 ; j < 9 ; j ++ ){
				if( game->Board[i][j].Belong == game->CurrentPlayer ){
					Sight_game.Board[8-i][8-j].Belong = JIBUN;
				}
				else if( game->Board[i][j].Belong == game->NonCurrentPlayer ){
					Sight_game.Board[8-i][8-j].Belong = TEKI;
				}
				else {
					Sight_game.Board[8-i][8-j].Belong = NASHI;
				}
				Sight_game.Board[8-i][8-j].PieceType
				= game->Board[i][j].PieceType;
				Sight_game.Board[8-i][8-j].isPromoted
				= game->Board[i][j].isPromoted;
			}
		}
/* Ǥ  =   ǤϤʤ = (CurrentPlayer) Ȥˤʤ */
		Sight_game.BlackPiecesInHand.Pawn = game->WhitePiecesInHand.Pawn;
		Sight_game.BlackPiecesInHand.Lance = game->WhitePiecesInHand.Lance;
		Sight_game.BlackPiecesInHand.Knight = game->WhitePiecesInHand.Knight;
		Sight_game.BlackPiecesInHand.Silver = game->WhitePiecesInHand.Silver;
		Sight_game.BlackPiecesInHand.Gold = game->WhitePiecesInHand.Gold;
		Sight_game.BlackPiecesInHand.Bishop = game->WhitePiecesInHand.Bishop;
		Sight_game.BlackPiecesInHand.Rook = game->WhitePiecesInHand.Rook;

		Sight_game.WhitePiecesInHand.Pawn = game->BlackPiecesInHand.Pawn;
		Sight_game.WhitePiecesInHand.Lance = game->BlackPiecesInHand.Lance;
		Sight_game.WhitePiecesInHand.Knight = game->BlackPiecesInHand.Knight;
		Sight_game.WhitePiecesInHand.Silver = game->BlackPiecesInHand.Silver;
		Sight_game.WhitePiecesInHand.Gold = game->BlackPiecesInHand.Gold;
		Sight_game.WhitePiecesInHand.Bishop = game->BlackPiecesInHand.Bishop;
		Sight_game.WhitePiecesInHand.Rook = game->BlackPiecesInHand.Rook;

		Sight_data.key = data->key;
		switch( data->key ){

			case MOVE:
				Sight_data.hand.move.sx = 8 - data->hand.move.sx;
				Sight_data.hand.move.sy = 8 - data->hand.move.sy;
				Sight_data.hand.move.dx = 8 - data->hand.move.dx;
				Sight_data.hand.move.dy = 8 - data->hand.move.dy;
				Sight_data.hand.move.willPromote
				= data->hand.move.willPromote;
				break;

			case DROP:
				Sight_data.hand.drop.dx = 8 - data->hand.drop.dx;
				Sight_data.hand.drop.dy = 8 - data->hand.drop.dy;
				Sight_data.hand.drop.piece = data->hand.drop.piece;
				break;

			default:
				break;
		}
	}


	switch( Sight_data.key ){
		case MOVE:
			if( Sight_game.Board[Sight_data.hand.move.sx]
			   [Sight_data.hand.move.sy].Belong != JIBUN ){
				return ILLEGAL;
			}
			flag = ExamineMove(	&Sight_game, &Sight_data, judge );

			if( flag != OK ){
				return flag;
			}
			else {
				return ExamineCheck( game, data, judge, &Sight_game,
									&Sight_data );
			}

		case DROP:
			if( Sight_game.Board[Sight_data.hand.drop.dx]
			   [Sight_data.hand.drop.dy].Belong == JIBUN ){
					judge->result = ILLEGAL;
					return ILLEGAL;
			}else if( Sight_game.Board[Sight_data.hand.drop.dx]
			   [Sight_data.hand.drop.dy].Belong == TEKI ){
					judge->result = VIOLATE;
					judge->violation = DROPON;
					return VIOLATE;
			}
			switch( Sight_data.hand.drop.piece ){

				case PAWN:
					for( i = 0 ; i < 9 ; i ++ ){
						if( Sight_game.Board[Sight_data.hand.drop.dx][i].Belong
						   == JIBUN && Sight_game.Board
						   [Sight_data.hand.drop.dx][i].PieceType == PAWN
						   && Sight_game.Board[Sight_data.hand.drop.dx][i]
							.isPromoted == 0 ){
							judge->result = ILLEGAL;
							return ILLEGAL;
						}			
					}
				case LANCE:
					if( Sight_data.hand.drop.dy == 0 ){
						judge->result = ILLEGAL;
						return ILLEGAL;
					}						
					return ExamineCheck( game, data, judge, &Sight_game,
									&Sight_data );


				case KNIGHT:
					if( Sight_data.hand.drop.dy < 2 ){
						judge->result = ILLEGAL;
						return ILLEGAL;
					}						
					return ExamineCheck( game, data, judge, &Sight_game,
									&Sight_data );
				default:
					return ExamineCheck( game, data, judge, &Sight_game,
									&Sight_data );
			}

		default:
			break;
	}
	return ILLEGAL;

} /* ExamineHandSub */


JUDGERESULT ExamineMove
(
	GAMEDATA *Sight_game,
	CLIENT_MESSAGE *Sight_data,
	JUDGE *judge
)
{

	if( Sight_data->hand.move.willPromote ){
		switch( (int) Sight_game->Board[Sight_data->hand.move.sx]
			   [Sight_data->hand.move.sy].Belong ){
			case JIBUN:
				if( Sight_data->hand.move.sy >2
				   && Sight_data->hand.move.dy >2 ){
					return ILLEGAL;
				}					
				break;

			case TEKI:
				if( Sight_data->hand.move.sy < 6
				   && Sight_data->hand.move.dy < 6 ){
					return ILLEGAL;
				}					
				break;

			default:
				return ILLEGAL;
		}
	}

	switch( Sight_game->Board[Sight_data->hand.move.sx]
		   [Sight_data->hand.move.sy].PieceType ){

		case PAWN:
			if( Sight_game->Board[Sight_data->hand.move.sx]
		   [Sight_data->hand.move.sy].isPromoted ){
				return ExamineMove_Gold( Sight_game, Sight_data, judge );
			}
			else{
				return ExamineMove_Pawn( Sight_game, Sight_data, judge );
			}

		case LANCE:
			if( Sight_game->Board[Sight_data->hand.move.sx]
		   [Sight_data->hand.move.sy].isPromoted ){
				return ExamineMove_Gold( Sight_game, Sight_data, judge );
			}
			else{
				return ExamineMove_Lance( Sight_game, Sight_data, judge );
			}

		case KNIGHT:
			if( Sight_game->Board[Sight_data->hand.move.sx]
		   [Sight_data->hand.move.sy].isPromoted ){
				return ExamineMove_Gold( Sight_game, Sight_data, judge );
			}
			else{
				return ExamineMove_Knight( Sight_game, Sight_data, judge );
			}

		case SILVER:
			if( Sight_game->Board[Sight_data->hand.move.sx]
		   [Sight_data->hand.move.sy].isPromoted ){
				return ExamineMove_Gold( Sight_game, Sight_data, judge );
			}
			else{
				return ExamineMove_Silver( Sight_game, Sight_data, judge );
			}

		case GOLD:
			return ExamineMove_Gold( Sight_game, Sight_data, judge );

		case BISHOP:
			if( Sight_game->Board[Sight_data->hand.move.sx]
		   [Sight_data->hand.move.sy].isPromoted ){
				return ExamineMove_Uma( Sight_game, Sight_data, judge );
			}
			else{
				return ExamineMove_Bishop( Sight_game, Sight_data, judge );
			}

		case ROOK:
			if( Sight_game->Board[Sight_data->hand.move.sx]
		   [Sight_data->hand.move.sy].isPromoted ){
				return ExamineMove_Ryu( Sight_game, Sight_data, judge );
			}
			else{
				return ExamineMove_Rook( Sight_game, Sight_data, judge );
			}

		case KING:
			return ExamineMove_King( Sight_game, Sight_data, judge );
		default:
			break;
		}

	return ILLEGAL;

} /* ExamineMove */

JUDGERESULT ExamineMove_Pawn
(	GAMEDATA *Sight_game,
 	CLIENT_MESSAGE *Sight_data,
	JUDGE *judge
)
{
	if( Sight_game->Board[Sight_data->hand.move.sx][Sight_data->hand.move.sy]
	   .Belong == Sight_game->Board[Sight_data->hand.move.dx]
	   [Sight_data->hand.move.dy].Belong || Sight_data->hand.move.sx
	   != Sight_data->hand.move.dx ){
			return ILLEGAL;
	}

	switch( (int) Sight_game->Board[Sight_data->hand.move.sx]
		   [Sight_data->hand.move.sy].Belong ){

		case JIBUN:
			if( Sight_data->hand.move.dy == 0
			   && Sight_data->hand.move.willPromote == 0 ){ 
				return ILLEGAL;
			}
			if( Sight_data->hand.move.sy - Sight_data->hand.move.dy == 1 ){
				return OK;
			}
			else{
				return ILLEGAL;
			}

		case TEKI:
			if( Sight_data->hand.move.dy == 8
			   && Sight_data->hand.move.willPromote == 0 ){
				return ILLEGAL;
			}
			if( Sight_data->hand.move.sy - Sight_data->hand.move.dy == -1 ){
				return OK;
			}
			else{
				return ILLEGAL;
			}

		default:
			break;
	}
	return ILLEGAL;
}

JUDGERESULT ExamineMove_Lance
(	GAMEDATA *Sight_game,
 	CLIENT_MESSAGE *Sight_data,
	JUDGE *judge
)
{
	int i;

	if( Sight_game->Board[Sight_data->hand.move.sx][Sight_data->hand.move.sy]
	   .Belong == Sight_game->Board[Sight_data->hand.move.dx]
	   [Sight_data->hand.move.dy].Belong || Sight_data->hand.move.sx
	   != Sight_data->hand.move.dx ){
			return ILLEGAL;
	}

	switch( (int)Sight_game->Board[Sight_data->hand.move.sx]
		   [Sight_data->hand.move.sy].Belong ){

		case JIBUN:
			if( Sight_data->hand.move.dy == 0
			   && Sight_data->hand.move.willPromote == 0 ){
				return ILLEGAL;
			}
			if( Sight_data->hand.move.sy > Sight_data->hand.move.dy ){
				for( i = Sight_data->hand.move.sy - 1 ;
					i > Sight_data->hand.move.dy	; i -- ){
					if( Sight_game->Board[Sight_data->hand.move.sx][i].Belong
					   == JIBUN ){
						return ILLEGAL;
					}
					if( Sight_game->Board[Sight_data->hand.move.sx][i].Belong
					   == TEKI ){
						judge->violation = JUMPOVER;
						return VIOLATE;
					}
				}
				return OK;
			}
			else{
				return ILLEGAL;
			}

		case TEKI:
			if( Sight_data->hand.move.dy == 0
			   && Sight_data->hand.move.willPromote == 8 ){
				return ILLEGAL;
			}
			if( Sight_data->hand.move.sy < Sight_data->hand.move.dy ){
				for( i = Sight_data->hand.move.sy + 1 ;
					i < Sight_data->hand.move.dy ; i ++ ){
					if( Sight_game->Board[Sight_data->hand.move.sx][i].Belong
					   == TEKI ){
						return ILLEGAL;
					}
					if( Sight_game->Board[Sight_data->hand.move.sx][i].Belong
					   == JIBUN ){
						judge->violation = JUMPOVER;
						return VIOLATE;
					}
				}
				return OK;
			}
			else{
				return ILLEGAL;
			}

		default:
			break;
	}

	return ILLEGAL;
}

JUDGERESULT ExamineMove_Knight
(	GAMEDATA *Sight_game,
 	CLIENT_MESSAGE *Sight_data,
	JUDGE *judge
)
{
	if( Sight_game->Board[Sight_data->hand.move.sx][Sight_data->hand.move.sy]
	   .Belong == Sight_game->Board[Sight_data->hand.move.dx]
	   [Sight_data->hand.move.dy].Belong ){
			return ILLEGAL;
	}
	switch( (int) (Sight_game->Board[Sight_data->hand.move.sx]
		   [Sight_data->hand.move.sy].Belong) ){

		case JIBUN:
		  Err("come to JIBUN.");
			if( Sight_data->hand.move.dy < 2
			   && Sight_data->hand.move.willPromote == 0 ){
				return ILLEGAL;
			}
			if( Sight_data->hand.move.sy - Sight_data->hand.move.dy == 2
			   && ( ( Sight_data->hand.move.sx - Sight_data->hand.move.dx == 1
				   && Sight_data->hand.move.sx != 0 )
			   || ( Sight_data->hand.move.sx - Sight_data->hand.move.dx == -1
				   && Sight_data->hand.move.sx != 8 ) ) ){
				return OK;
			}
			else{
				return ILLEGAL;
			}

		case TEKI:
		  Err("come to TEKI.");
			if( Sight_data->hand.move.dy > 6
			   && Sight_data->hand.move.willPromote == 0 ){
				return ILLEGAL;
			}
			if( Sight_data->hand.move.sy - Sight_data->hand.move.dy == -2
			   && ( ( Sight_data->hand.move.sx - Sight_data->hand.move.dx == 1
				   && Sight_data->hand.move.sx != 0 )
				   || ( Sight_data->hand.move.sx - Sight_data->hand.move.dx == -1
				   && Sight_data->hand.move.sx != 8 ) ) ){
					return OK;
			}
			else{
				return ILLEGAL;
			}

		default:
		  Err("come to default.");
			break;
	}
	return ILLEGAL;
}

JUDGERESULT ExamineMove_Silver
(	GAMEDATA *Sight_game,
 	CLIENT_MESSAGE *Sight_data,
	JUDGE *judge
)
{
	if( Sight_game->Board[Sight_data->hand.move.sx][Sight_data->hand.move.sy]
	   .Belong == Sight_game->Board[Sight_data->hand.move.dx]
	   [Sight_data->hand.move.dy].Belong || Sight_data->hand.move.dx > 8
	   || Sight_data->hand.move.dx < 0 || Sight_data->hand.move.dy > 8
	   || Sight_data->hand.move.dx < 0 ){
			return ILLEGAL;
	}

	if( abs( Sight_data->hand.move.sx - Sight_data->hand.move.dx ) == 1
	   && abs( Sight_data->hand.move.sy - Sight_data->hand.move.dy ) == 1 ){
		return OK;
	}
	if( Sight_data->hand.move.sx != Sight_data->hand.move.dx ){
		return ILLEGAL;
	}


	switch( (int)Sight_game->Board[Sight_data->hand.move.sx]
		   [Sight_data->hand.move.sy].Belong ){

		case JIBUN:
			if( Sight_data->hand.move.sy - Sight_data->hand.move.dy == 1 ){
					return OK;
			}
			else{
				return ILLEGAL;
			}

		case TEKI:
			if( Sight_data->hand.move.sy - Sight_data->hand.move.dy == -1 ){
					return OK;
			}
			else{
				return ILLEGAL;
			}

		default:
			break;
	}
	return ILLEGAL;
}

JUDGERESULT ExamineMove_Gold
(	GAMEDATA *Sight_game,
 	CLIENT_MESSAGE *Sight_data,
	JUDGE *judge
)
{
	if( Sight_game->Board[Sight_data->hand.move.sx][Sight_data->hand.move.sy]
	   .Belong == Sight_game->Board[Sight_data->hand.move.dx]
	   [Sight_data->hand.move.dy].Belong || Sight_data->hand.move.dx > 8
	   || Sight_data->hand.move.dx < 0 || Sight_data->hand.move.dy > 8
	   || Sight_data->hand.move.dx < 0 ){
			return ILLEGAL;
	}

	switch( (int)Sight_game->Board[Sight_data->hand.move.sx]
		   [Sight_data->hand.move.sy].Belong ){

		case JIBUN:
			if( ( Sight_data->hand.move.sy - Sight_data->hand.move.dy == 1
			   && abs(Sight_data->hand.move.sx - Sight_data->hand.move.dx) < 2 )
			   ||
			   ( Sight_data->hand.move.sy == Sight_data->hand.move.dy
			   && abs(Sight_data->hand.move.sx - Sight_data->hand.move.dx) == 1 )
			   ||
			   ( Sight_data->hand.move.sy - Sight_data->hand.move.dy == -1
			   && Sight_data->hand.move.sx == Sight_data->hand.move.dx ) ){
				return OK;
			}
			else{
				return ILLEGAL;
			}

		case TEKI:
			if( ( Sight_data->hand.move.sy - Sight_data->hand.move.dy == -1
			   && abs(Sight_data->hand.move.sx - Sight_data->hand.move.dx) < 2 )
			   ||
			   ( Sight_data->hand.move.sy == Sight_data->hand.move.dy
			   && abs(Sight_data->hand.move.sx - Sight_data->hand.move.dx) == 1 )
			   ||
			   ( Sight_data->hand.move.sy - Sight_data->hand.move.dy == 1
			   && Sight_data->hand.move.sx == Sight_data->hand.move.dx ) ){
				return OK;
			}
			else{
				return ILLEGAL;
			}

		default:
			break;
	}
	return ILLEGAL;
}

JUDGERESULT ExamineMove_King
(	GAMEDATA *Sight_game,
 	CLIENT_MESSAGE *Sight_data,
	JUDGE *judge
)
{
	if( Sight_game->Board[Sight_data->hand.move.sx][Sight_data->hand.move.sy]
	   .Belong == Sight_game->Board[Sight_data->hand.move.dx]
	   [Sight_data->hand.move.dy].Belong || Sight_data->hand.move.dx > 8
	   || Sight_data->hand.move.dx < 0 || Sight_data->hand.move.dy > 8
	   || Sight_data->hand.move.dx < 0 ){
			return ILLEGAL;
	}

	if( abs(Sight_data->hand.move.sx - Sight_data->hand.move.dx) < 2
		&& abs(Sight_data->hand.move.sy - Sight_data->hand.move.dy) < 2 ){
		return OK;
	}
	return ILLEGAL;
}

JUDGERESULT ExamineMove_Bishop
(
 	GAMEDATA *Sight_game,
 	CLIENT_MESSAGE *Sight_data,
	JUDGE *judge
)
{
	int i,deltax,deltay;

	if( Sight_game->Board[Sight_data->hand.move.sx][Sight_data->hand.move.sy]
	   .Belong == Sight_game->Board[Sight_data->hand.move.dx]
	   [Sight_data->hand.move.dy].Belong
	   || abs(Sight_data->hand.move.sx - Sight_data->hand.move.dx )
	   != abs(Sight_data->hand.move.sy - Sight_data->hand.move.dy )
	   || Sight_data->hand.move.dx > 8 || Sight_data->hand.move.dx < 0
	   || Sight_data->hand.move.dy > 8 || Sight_data->hand.move.dx < 0 ){
			return ILLEGAL;
	}

	deltax = ( Sight_data->hand.move.dx - Sight_data->hand.move.sx )
			/ abs(Sight_data->hand.move.sx - Sight_data->hand.move.dx );
	deltay = ( Sight_data->hand.move.dy - Sight_data->hand.move.sy )
			/ abs(Sight_data->hand.move.sy - Sight_data->hand.move.dy );

	for( i = 1 ; i < abs(Sight_data->hand.move.sx - Sight_data->hand.move.dx )
		; i++ ){
		if( Sight_game->Board[Sight_data->hand.move.sx + deltax * i]
		   [Sight_data->hand.move.sy + deltay * i].Belong
		   == Sight_game->Board[Sight_data->hand.move.sx]
		   [Sight_data->hand.move.sy].Belong ){
			return ILLEGAL;
		}
		else if( Sight_game->Board[Sight_data->hand.move.sx + deltax * i]
		   [Sight_data->hand.move.sy + deltay * i].Belong != NASHI
			){
			judge->violation = JUMPOVER;
			return VIOLATE;
		}
	}

	return OK;

}

JUDGERESULT ExamineMove_Rook
(
	GAMEDATA *Sight_game,
 	CLIENT_MESSAGE *Sight_data,
	JUDGE *judge
)
{
	int i,delta;

	if( Sight_game->Board[Sight_data->hand.move.sx][Sight_data->hand.move.sy]
	   .Belong == Sight_game->Board[Sight_data->hand.move.dx]
	   [Sight_data->hand.move.dy].Belong
	   || ( Sight_data->hand.move.sx != Sight_data->hand.move.dx
		   && Sight_data->hand.move.sy != Sight_data->hand.move.dy )
	   || Sight_data->hand.move.dx > 8 || Sight_data->hand.move.dx < 0
	   || Sight_data->hand.move.dy > 8 || Sight_data->hand.move.dy < 0 ){
			return ILLEGAL;
	}

	if(  Sight_data->hand.move.sx == Sight_data->hand.move.dx ){

		delta = ( Sight_data->hand.move.dy - Sight_data->hand.move.sy )
		 / abs(Sight_data->hand.move.sy - Sight_data->hand.move.dy );

		for( i = 1 ; i < abs(Sight_data->hand.move.sy
							 - Sight_data->hand.move.dy ) ; i++ ){
			if( Sight_game->Board[Sight_data->hand.move.sx]
			   [Sight_data->hand.move.sy + delta * i].Belong
			   == Sight_game->Board[Sight_data->hand.move.sx]
			   [Sight_data->hand.move.sy].Belong ){
				return ILLEGAL;
			}
			if( Sight_game->Board[Sight_data->hand.move.sx]
			   [Sight_data->hand.move.sy + delta * i].Belong != NASHI ){
				judge->violation = JUMPOVER;
				return VIOLATE;
			}
		}
	}

	if(  Sight_data->hand.move.sy == Sight_data->hand.move.dy ){

		delta = ( Sight_data->hand.move.dx - Sight_data->hand.move.sx )
		 / abs(Sight_data->hand.move.sx - Sight_data->hand.move.dx );

		for( i = 1 ; i < abs(Sight_data->hand.move.sx
							 - Sight_data->hand.move.dx ) ; i++ ){
			if( Sight_game->Board[Sight_data->hand.move.sx + delta * i]
			   [Sight_data->hand.move.sy].Belong
			   == Sight_game->Board[Sight_data->hand.move.sx]
			   [Sight_data->hand.move.sy].Belong ){
				return ILLEGAL;
			}
			if( Sight_game->Board[Sight_data->hand.move.sx + delta * i]
			   [Sight_data->hand.move.sy].Belong != NASHI ){
				judge->violation = JUMPOVER;
				return VIOLATE;
			}
		}
	}

	return OK;

}

JUDGERESULT ExamineMove_Uma
(
	GAMEDATA *Sight_game,
 	CLIENT_MESSAGE *Sight_data,
	JUDGE *judge
)
{
	if(  ExamineMove_King( Sight_game, Sight_data, judge ) == OK )return OK;

	return ExamineMove_Bishop( Sight_game, Sight_data, judge );
}

JUDGERESULT ExamineMove_Ryu
(
	GAMEDATA *Sight_game,
 	CLIENT_MESSAGE *Sight_data,
	JUDGE *judge
)
{
	if(  ExamineMove_King( Sight_game, Sight_data, judge ) == OK )return OK;

	return ExamineMove_Rook( Sight_game, Sight_data, judge );
}

/* ExamineMove_honya ޤ */






JUDGERESULT ExamineCheck
(
	GAMEDATA *game,
	const CLIENT_MESSAGE *data,
	JUDGE *judge,
	GAMEDATA *Sight_game,
	CLIENT_MESSAGE *Sight_data
)
{
	int i,j,k,l,flag=0;
	int King_x=0,King_y=0;
	CLIENT_MESSAGE tmp_message;


	/* äưƹͤƤߤ褦 */
	switch( Sight_data->key ){

		case MOVE:
			Sight_game->Board[Sight_data->hand.move.dx]
			[Sight_data->hand.move.dy].Belong
			= Sight_game->Board[Sight_data->hand.move.sx]
			[Sight_data->hand.move.sy].Belong;

			Sight_game->Board[Sight_data->hand.move.dx]
			[Sight_data->hand.move.dy].PieceType
			= Sight_game->Board[Sight_data->hand.move.sx]
			[Sight_data->hand.move.sy].PieceType;

			Sight_game->Board[Sight_data->hand.move.dx]
			[Sight_data->hand.move.dy].isPromoted
			= Sight_game->Board[Sight_data->hand.move.sx]
			[Sight_data->hand.move.sy].isPromoted;

			if( Sight_data->hand.move.willPromote ){
				Sight_game->Board[Sight_data->hand.move.dx]
				[Sight_data->hand.move.dy].isPromoted = 1;
			}

			Sight_game->Board[Sight_data->hand.move.sx]
			[Sight_data->hand.move.sy].Belong	= NASHI;

			Sight_game->Board[Sight_data->hand.move.sx]
			[Sight_data->hand.move.sy].PieceType = NONE;
			break;

		case DROP:
			Sight_game->Board[Sight_data->hand.drop.dx]
			[Sight_data->hand.drop.dy].Belong = JIBUN;
			Sight_game->Board[Sight_data->hand.drop.dx]
			[Sight_data->hand.drop.dy].PieceType
			= Sight_data->hand.drop.piece;
			Sight_game->Board[Sight_data->hand.drop.dx]
			[Sight_data->hand.drop.dy].isPromoted = 0;
			break;

		default:
			break;
	}	



	/* ̤˲ Ϥ뤫*/
	for( i = 0 ; i < 9 ; i ++ ){
		for( j = 0 ; j < 9 ; j ++ ){
			if( Sight_game->Board[i][j].PieceType == KING
			   && Sight_game->Board[i][j].Belong == JIBUN ){
				King_x = i;
				King_y = j;
			}
		}
	}
	for( i = 0 ; i < 9 ; i ++ ){
		for( j = 0 ; j < 9 ; j ++ ){
			tmp_message.key = MOVE;
			tmp_message.hand.move.sx = i;
			tmp_message.hand.move.sy = j;
			tmp_message.hand.move.dx = King_x;
			tmp_message.hand.move.dy = King_y;
			tmp_message.hand.move.willPromote = 0;
			if( ExamineMove( Sight_game, &tmp_message, judge ) == OK ){
				judge->violation = CHECK;
				return VIOLATE;
			}
			tmp_message.hand.move.willPromote = 1;
			if( ExamineMove( Sight_game, &tmp_message, judge ) == OK ){
				judge->violation = CHECK;
				return VIOLATE;
			}
		}
	}



	/* Ũϵͤ*/
	flag = 0;
	for( i = 0 ; i < 9 ; i ++ ){
		for( j = 0 ; j < 9 ; j ++ ){
			if( Sight_game->Board[i][j].Belong == TEKI ){
				for( k = 0 ; k < 9 ; k ++ ){
					for( l = 0 ; l < 9 ; l ++ ){
						tmp_message.key = MOVE;
						tmp_message.hand.move.sx = i;
						tmp_message.hand.move.sy = j;
						tmp_message.hand.move.dx = k;
						tmp_message.hand.move.dy = l;
						tmp_message.hand.move.willPromote = 0;
						if( ExamineMove( Sight_game, &tmp_message, judge )
						   == OK ){
							if( ExamineCheck_Sub( Sight_game, &tmp_message )
							   == 0 ){
								flag = 1;
							}
						}
					tmp_message.hand.move.willPromote = 1;
						if( ExamineMove( Sight_game, &tmp_message, judge )
						   == OK ){
							if( ExamineCheck_Sub( Sight_game, &tmp_message )
							   == 0 ){
								flag = 1;
							}
						}
					}
				}
			}
		}
	}
	if( Sight_game->WhitePiecesInHand.Silver
	   || Sight_game->WhitePiecesInHand.Gold
	   || Sight_game->WhitePiecesInHand.Bishop
	   || Sight_game->WhitePiecesInHand.Rook ){
		for( i = 0 ; i < 9 ; i ++ ){
			for( j = 0 ; j < 9 ; j ++ ){
				if( Sight_game->Board[i][j].Belong == NASHI ){
					tmp_message.key = DROP;
					tmp_message.hand.drop.dx = i;
					tmp_message.hand.drop.dy = j;
					if( ExamineCheck_Sub( Sight_game, &tmp_message ) == 0 ){
						flag = 1;
					}
				}
			}
		}
	}
	else if( Sight_game->WhitePiecesInHand.Lance ){
		for( i = 0 ; i < 9 ; i ++ ){
			for( j = 0 ; j < 8 ; j ++ ){
				if( Sight_game->Board[i][j].Belong == NASHI ){
					tmp_message.key = DROP;
					tmp_message.hand.drop.dx = i;
					tmp_message.hand.drop.dy = j;
					if( ExamineCheck_Sub( Sight_game, &tmp_message ) == 0 ){
						flag = 1;
					}
				}
			}
		}
	}
	else if( Sight_game->WhitePiecesInHand.Knight ){
		for( i = 0 ; i < 9 ; i ++ ){
			for( j = 0 ; j < 7 ; j ++ ){
				if( Sight_game->Board[i][j].Belong == NASHI ){
					tmp_message.key = DROP;
					tmp_message.hand.drop.dx = i;
					tmp_message.hand.drop.dy = j;
					if( ExamineCheck_Sub( Sight_game, &tmp_message ) == 0 ){
						flag = 1;
					}
				}
			}
		}
	}
	else if( Sight_game->WhitePiecesInHand.Pawn ){
		for( i = 0 ; i < 9 ; i ++ ){
			for( j = 0 ; j < 8 ; j ++ ){
				if( Sight_game->Board[i][j].PieceType == PAWN
				   && Sight_game->Board[i][j].Belong == TEKI
				   && Sight_game->Board[i][j].isPromoted == 0 ){
					break;
				}
			}
			if( j < 8 )continue;
			for( j = 0 ; j < 8 ; j ++ ){
				if( Sight_game->Board[i][j].Belong == NASHI ){
					tmp_message.key = DROP;
					tmp_message.hand.drop.dx = i;
					tmp_message.hand.drop.dy = j;
					if( ExamineCheck_Sub( Sight_game, &tmp_message ) == 0 ){
						flag = 1;
					}
				}
			}
		}
	}
	if( flag == 0 ){
		if( data->key == DROP && data->hand.drop.piece == PAWN ){
			judge->violation = CHECKMATEBYDROPPINGPAWN;			
			return VIOLATE;
		}
		else{
			judge->endstate.winner = game->CurrentPlayer;
			judge->endstate.reason = CHECKMATED;
			if( data->key == MOVE && game->Board[data->hand.move.dx]
			   [data->hand.move.dy].Belong == game->NonCurrentPlayer ){
				judge->isCaptured = 1;
				judge->capture.piece = game->Board[data->hand.move.dx]
				[data->hand.move.dy].PieceType;
				judge->capture.x = data->hand.move.dx;
				judge->capture.y = data->hand.move.dy;
				judge->capture.hasPromoted = game->Board[data->hand.move.dx]
				[data->hand.move.dy].isPromoted;
			}
			return END;
		}
	}


	/* Ṳ̃˲ Ϥä*/
	for( i = 0 ; i < 9 ; i ++ ){
		for( j = 0 ; j < 9 ; j ++ ){
			if( Sight_game->Board[i][j].PieceType == KING
			   && Sight_game->Board[i][j].Belong == TEKI ){
				King_x = i;
				King_y = j;
			}
		}
	}
	for( i = 0 ; i < 9 ; i ++ ){
		for( j = 0 ; j < 9 ; j ++ ){
			tmp_message.key = MOVE;
			tmp_message.hand.move.sx = i;
			tmp_message.hand.move.sy = j;
			tmp_message.hand.move.dx = King_x;
			tmp_message.hand.move.dy = King_y;
			tmp_message.hand.move.willPromote = 0;
			if( ExamineMove( Sight_game, &tmp_message, judge ) == OK ){
				judge->isCheck = 1;
				return OK;
			}
			tmp_message.hand.move.willPromote = 1;
			if( ExamineMove( Sight_game, &tmp_message, judge ) == OK ){
				judge->isCheck = 1;
				return OK;
			}
		}
	}

	return OK;

} /* ExamineCheck */


int ExamineCheck_Sub
(
	GAMEDATA *Sight_game,
	CLIENT_MESSAGE *tmp_data
)
{

	int i,j;
	int King_x=0,King_y=0;
	GAMEDATA tmp_game;
	CLIENT_MESSAGE tmp_message;
	JUDGE judge_dammy;


	for( i = 0 ; i < 9 ; i ++ ){
		for( j = 0 ; j < 9 ; j ++ ){
			 tmp_game.Board[i][j].PieceType = Sight_game->Board[i][j]
			 .PieceType;
			 tmp_game.Board[i][j].Belong = Sight_game->Board[i][j].Belong;
			 tmp_game.Board[i][j].isPromoted = Sight_game->Board[i][j]
			 .isPromoted;
		 }
	}

	switch( tmp_data->key ){

		case MOVE:
			tmp_game.Board[tmp_data->hand.move.dx][tmp_data->hand.move.dy]
			.Belong	= TEKI;
			tmp_game.Board[tmp_data->hand.move.dx][tmp_data->hand.move.dy]
			.PieceType = tmp_game.Board[tmp_data->hand.move.sx]
			  [tmp_data->hand.move.sy].PieceType;
			tmp_game.Board[tmp_data->hand.move.sx][tmp_data->hand.move.sy]
			.Belong	= NASHI;
			tmp_game.Board[tmp_data->hand.move.sx][tmp_data->hand.move.sy]
			.PieceType = NONE;
			break;

		case DROP:
			tmp_game.Board[tmp_data->hand.drop.dx][tmp_data->hand.drop.dy]
			.Belong	= TEKI;
			tmp_game.Board[tmp_data->hand.drop.dx][tmp_data->hand.drop.dy]
			.PieceType = ROOK;
			break;

		default:
			break;
	}	

	/* Ṳ̃˲ Ϥä*/
	for( i = 0 ; i < 9 ; i ++ ){
		for( j = 0 ; j < 9 ; j ++ ){
			if( tmp_game.Board[i][j].PieceType == KING
			   && tmp_game.Board[i][j].Belong == TEKI ){
				King_x = i;
				King_y = j;
			}
		}
	}
	for( i = 0 ; i < 9 ; i ++ ){
		for( j = 0 ; j < 9 ; j ++ ){
			tmp_message.key = MOVE;
			tmp_message.hand.move.sx = i;
			tmp_message.hand.move.sy = j;
			tmp_message.hand.move.dx = King_x;
			tmp_message.hand.move.dy = King_y;
			tmp_message.hand.move.willPromote = 0;
			if( ExamineMove( &tmp_game, &tmp_message, &judge_dammy ) == OK ){
				return 1;
			}
			tmp_message.hand.move.willPromote = 1;
			if( ExamineMove( &tmp_game, &tmp_message, &judge_dammy ) == OK ){
				return 1;
			}
		}
	}
	return 0;

} /* ExamineCheck_Sub */


