/* ** main.c ************************************* */
/* Main routine of TUITATED. */
/* *********************************************** */

/* -- Include File ------------------------------- */
#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <dirent.h>
#include <sys/types.h>

#include "tuitated.h"

int             ApplySocket;
unsigned short  port;

int             InDebugMode;
int             InRandomMode;

int             isAccessLogFileSpecified;
int             isServiceLogFileSpecified;
int             isErrorLogFileSpecified;

char            AccessLogFile[512];
char            ServiceLogFile[512];
char            ErrorLogFile[512];

char            NotationFile[512];
char            ExportDir[512];

int             ExpireDays;

int             ApplySocketAlreadyCreated;
int             ShouldReopenApplySocket;

static char     ConfigFile[512];
static int      isPortSpecifiedInCommandLine;

/* -- Protype Declerations  ---------------------- */
int main( int argc, char **argv );
void SetDefaultParames( void );
int ParseConfigFile( void );
void ParseArg( int argc, char **argv );
int TryToOpenFile( char *fliename );
int TryToOpenDir( char *dirname );
void Init( void ); 	/* init.c */
void Apply( void );	/* apply.c */

/* ** Func main ********************************** */
/* main routine of tuitate */
int main( int argc, char **argv )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
    SetDefaultParames();
    ParseArg( argc, argv );
    ParseConfigFile();
    Init();
    ServiceLog( "Started ----- " VERSION " --------- " );
    Apply(); /* loop */

    /* not reach here */
    return 0;
} /* main */


/* ** Func SetDefaultParams ********************** */
/*  */
void SetDefaultParames( void )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
    /* PORT is defined in config.h */
    port = PORT;
    InDebugMode = FALSE;
    InRandomMode = FALSE;
    /* CONFIGFILE is defined in Makefile */
    strcpy( ConfigFile, CONFIGFILE );
    /* NOTATIONFILE is defined in Makefile*/
    strcpy( NotationFile, NOTATIONFILE );
    /* EXPORTDIR is defined in Makefile */
    strcpy( ExportDir, EXPORTDIR );

    isAccessLogFileSpecified = FALSE;
    isServiceLogFileSpecified = FALSE;
    isErrorLogFileSpecified = FALSE;

    isPortSpecifiedInCommandLine = FALSE;

    /* DEFAULT_EXPIREDAYS is defined in config.h */
    ExpireDays = DEFAULT_EXPIREDAYS;

    ApplySocketAlreadyCreated = FALSE;
    ShouldReopenApplySocket = FALSE;
} /* SetDefaultParams */

/* ** Func ParseConfigFile ******************************* */
/*
 * Read and parse tuitate config file.
 * File is in global char array 'ConfigFile'.
 * Return 0 on success.
 * Return 1 on failure.
 */
int ParseConfigFile( void )
{
/* -- Local Data --------------------------------- */
    FILE *fp;
    char buf[256];
    char *p;
/* -- Operation ---------------------------------- */
    fp = fopen( ConfigFile, "r" );
    if( fp == NULL ) {
        Err( "Cannot open configuration file '%s'.", ConfigFile );
        return 1;
    }
    isAccessLogFileSpecified = FALSE;
    isServiceLogFileSpecified = FALSE;
    isErrorLogFileSpecified = FALSE;

    while( fgets( buf, 256, fp ) != NULL ) {
        if( buf[0] == '#' ) continue;
        p = strtok( buf, " \t\n" );
        if( p == NULL ) continue;
        if( !strcmp( buf, "PORT" ) ) {
            p = strtok( NULL, " \t\n" );
            if( p != NULL ) {
                if( !isPortSpecifiedInCommandLine ) {
                    short newport;
                    newport = atoi( p );
                    if( newport != port ) {
                        ShouldReopenApplySocket = TRUE;
                        port = newport;
                    }
                }
            }
        } else if( !strcmp( buf, "ACCESSLOG" ) ) {
            p = strtok( NULL, " \t\n" );
            if( p != NULL ) {
                if( !TryToOpenFile( p ) ) {
                    strcpy( AccessLogFile, p );
                    isAccessLogFileSpecified = TRUE;
                } else {
                     Err( "Specified ACCESSLOG file is not apropriate." );
                }
            }
        } else if( !strcmp( buf, "SERVICELOG" ) ) {
            p = strtok( NULL, " \t\n" );
            if( p != NULL ) {
                if( !TryToOpenFile( p ) ) {
                    strcpy( ServiceLogFile, p );
                    isServiceLogFileSpecified = TRUE;
                } else {
                     Err( "Specified SERVICELOG file is not apropriate." );
                }
            }
        } else if( !strcmp( buf, "ERRORLOG" ) ) {
            p = strtok( NULL, " \t\n" );
            if( p != NULL ) {
                if( !TryToOpenFile( p ) ) {
                    isErrorLogFileSpecified = TRUE;
                    strcpy( ErrorLogFile, p );
                } else {
                     Err( "Specified ERRORLOG file is not apropriate." );
                }
            }
        } else if( !strcmp( buf, "NOTATIONFILE" ) ) {
            p = strtok( NULL, " \t\n" );
            if( p != NULL ) {
                if( !TryToOpenFile( p ) ) {
                    strcpy( NotationFile, p );
                } else {
                     Err( "Specified NOTATIONFILE is not apropriate." );
                }
            }
        } else if( !strcmp( buf, "EXPORTDIR" ) ) {
            p = strtok( NULL, " \t\n" );
            if( p != NULL ) {
                if( !TryToOpenDir( p ) ) {
                    strcpy( ExportDir, p );
                } else {
                     Err( "Specified EXPORTDIR is not apropriate." );
                }
            }
        } else if( !strcmp( buf, "EXPIREDAYS" ) ) {
            p = strtok( NULL, " \t\n" );
            if( p != NULL ) {
                if( !isPortSpecifiedInCommandLine ) {
                    ExpireDays = atoi( p );
                }
            }
        }
    }
    fclose( fp );
    return 0;
} /* ParseConfigFile */

/* ** Func ParseArg ****************************** */
/* Parse Argument */
void ParseArg( int argc, char **argv )
{
/* -- Local Data --------------------------------- */
    int i;
    int ExpectPortNum = FALSE;
    int ExpectConfigFileName = FALSE;
/* -- Operation ---------------------------------- */
    InDebugMode = FALSE;
    InRandomMode = TRUE;
    for( i = 1; i < argc; i++ ) {
        if( ExpectPortNum ) {
            port = atoi( argv[i] );
            isPortSpecifiedInCommandLine = TRUE;
            ExpectPortNum = FALSE;
        }
        if( ExpectConfigFileName ) {
            strcpy( ConfigFile, argv[i] );
            fprintf( stderr, "Specified config file is %s...\n", ConfigFile );
            ExpectConfigFileName = FALSE;
        }
        if( argv[i][0] == '-' ) {
            switch( argv[i][1] ) {
              case 'd':
                InDebugMode = TRUE;
                break;
              case 'h':
                printf( "Usage: tuitated [switches]\n"
                        "switches:\n"
                        "   -d           debug mode\n"
                        "   -h           display help\n"
                        "   -f conffile  specify configuration "
                                         "file with absolote path\n"
                        "   -n           decide black and white NOT randomly\n"
                        "   -p port      set port number to use\n"
                    );
                exit( 0 );
                break;
              case 'p':
                ExpectPortNum = TRUE;
                break;
              case 'n':
                InRandomMode = FALSE;
                break;
              case 'f':
                ExpectConfigFileName = TRUE;
                break;
              default:
                break;
            }
        }
    }
    return;
} /* ParseArg */

/* ** Func TryToOpenFile ************************* */
/*
 * Tru to open given file.
 * return 0 --- success.
 * return 1 --- fail.
 */
int TryToOpenFile( char *filename )
{
/* -- Local Data --------------------------------- */
    FILE *fp;
/* -- Operation ---------------------------------- */
    fp = fopen( filename, "a" );
    if( fp == NULL ) {
        return 1;
        Err( "fopen -- %s.", filename );
    }
    fclose( fp );
    return 0;
} /* TryToOpenFile */

/* ** Func TryToOpenDir ************************* */
/*
 * Tru to open given directory.
 * return 0 --- success.
 * return 1 --- fail.
 */
int TryToOpenDir( char *dirname )
{
/* -- Local Data --------------------------------- */
    DIR *dp;
/* -- Operation ---------------------------------- */
    dp = opendir( dirname );
    if( dp == NULL ) {
        return 1;
        Err( "opendir. -- %s", dirname );
    }
    closedir( dp );
    return 0;
} /* TryToOpenDir */







