/* ** notation.c ********************************** */
/* Routines to handle notation data. */
/* ************************************************ */

/* -- Include File ------------------------------- */
#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <unistd.h>
#include <sys/types.h>
#include <dirent.h>
#include <sys/stat.h>
#include "sendmessage.h"
#include "tuitated.h"

/* -- Global Data -------------------------------- */
NOTATIONDATA *NotationList;
extern char NotationFile[];
extern int ExpireDays;

/* -- Prototype Declerations  -------------------- */
void InitNotationList( void );
void UpdateNotationList( void );
void CleanupNotationList( void );
int NotationNum( void );
int CompareNotation( const void *a, const void *b );
int NewestNotationID( void );
int GetNumberFromString( const char *str );

/* ** Func InitNotationList ********************** */
/* Initialize Notation List */
void InitNotationList( void )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
    NotationList = NULL;
} /* InitNotationList */


/* ** Func UpdateNotationList ******************** */
/* Examin Notation Directory and update List data. */
void UpdateNotationList( void )
{
/* -- Local Data --------------------------------- */
    FILE *fp;
    char buf[1024];
    char tmp[1024];
    char *p;
    NOTATIONDATA *cur;
    int StillSearchNotation = TRUE;
/* -- Operation ---------------------------------- */
    if( NotationList != NULL ) {
        CleanupNotationList();
    }
    cur = NULL;
    fp = fopen( NotationFile, "r" );
    if( fp == NULL ) {
        Err( "Cannot Open Notation File -- %s.", NotationFile );
        return;
    }
    while( StillSearchNotation ) {
        int i;
        NOTATIONDATA *notation;
        /* Search notations in NOTATION FILE */
        while( 1 ) { 
            if( fgets( buf, 256 + 1, fp ) == NULL ) {
                StillSearchNotation = FALSE;
                break;
            }   
            if( !strncmp( buf, MAGIC, strlen( MAGIC ) ) ) { /* check MAGIC */
                break;
            }
        }
        if( !StillSearchNotation ) {
            break;
        }
        /* Allocate notation data for list. */
        notation = malloc( sizeof( NOTATIONDATA ) );
        if( notation == NULL ) {
            Err( "malloc." );
            break;
        }
        notation->Next = NULL;
        /* ID */
        notation->ID = GetNumberFromString( buf );
        /* SAVED or not SAVED? */
        if( strstr( buf, STR_SAVE ) == NULL ) 
            notation->saved = FALSE;
        else
            notation->saved = TRUE;
        /* Begin time */
        if( fgets( buf, 256 + 1, fp ) == NULL ) {
            Err( "fgets." );
            free( notation );
            break;
        }
        for( i = 0; isdigit( buf[i] ); i++ ) tmp[i] = buf[i];
        tmp[i] = '\0';
        notation->Begin = atol( tmp );
        /* End time */
        if( fgets( buf, 256 + 1, fp ) == NULL ) {
            Err( "fgets." );
            free( notation );
            break;
        }
        for( i = 0; isdigit( buf[i] ); i++ ) tmp[i] = buf[i];
        tmp[i] = '\0';
        notation->End = atol( tmp );

        /* Name and Address */
        if( fgets( buf, 256 + 5, fp ) == NULL ) {
            Err( "fgets." );
            free( notation );
            break;
        }

        for( i = 0; isgraph( buf[i] ); i++ ) tmp[i] = buf[i];
        tmp[i] = '\0';
        p = buf + i + 1;
        notation->BlackName = malloc( strlen( tmp ) + 1 );
        if( notation->BlackName == NULL ) {
            Err( "malloc." );
            free( notation );
            fclose( fp );
            return;
        }
        strcpy( notation->BlackName, tmp );

        for( i = 0; isgraph( p[i] ); i++ ) tmp[i] = p[i];
        tmp[i] = '\0';
        p += i + 1;
        notation->BlackAddress = malloc( strlen( tmp ) + 1 );
        if( notation->BlackAddress == NULL ) {
            Err( "malloc." );
            free( notation->BlackName );
            free( notation );
            fclose( fp );
        }
        strcpy( notation->BlackAddress, tmp );

        for( i = 0; isgraph( p[i] ); i++ ) tmp[i] = p[i];
        tmp[i] = '\0';
        p += i + 1;
        notation->WhiteName = malloc( strlen( tmp ) + 1 );
        if( notation->WhiteName == NULL ) {
            Err( "malloc." );
            free( notation->BlackName );
            free( notation->BlackAddress );
            free( notation );
            fclose( fp );
        }
        strcpy( notation->WhiteName, tmp );
            
        for( i = 0; isgraph( p[i] ); i++ ) tmp[i] = p[i];
        tmp[i] = '\0';
        p += i + 1;
        notation->WhiteAddress = malloc( strlen( tmp ) + 1 );
        if( notation->WhiteAddress == NULL ) {
            Err( "malloc." );
            free( notation->BlackName );
            free( notation->BlackAddress );
            free( notation->WhiteName );
            free( notation );
            fclose( fp );
        }
        strcpy( notation->WhiteAddress, tmp );

        /* Save Position of File to reread */
        notation->NotationPos = ftell( fp );

        /* Add prepared Notation data to List. */
        if( cur == NULL ) {
            NotationList = notation;
            cur = notation;
        } else {
            cur->Next = notation;
            cur = notation;
        }
    }
    fclose( fp );
} /* UdpateNotationList */

/* ** Func CompareNotation *************************** */
/*
 * Comapre Notations according to their END TIME.
 * For sorting of Notation List.
 */
int CompareNotation( const void *a, const void *b )
{
/* -- Local Data --------------------------------- */
    NOTATIONDATA *NotationA;
    NOTATIONDATA *NotationB;
/* -- Operation ---------------------------------- */
    NotationA = *(NOTATIONDATA **)a;
    NotationB = *(NOTATIONDATA **)b;
    if( NotationA->End < NotationB->End ) return -1;
    if( NotationA->End > NotationB->End ) return 1;
    return 0;
} /* CompareNotation */

/* ** Func CleanupNotationList ******************** */
/* Remove all data of Notation List. */
void CleanupNotationList( void )
{
/* -- Local Data --------------------------------- */
    NOTATIONDATA *cur;
    NOTATIONDATA *prev;
/* -- Operation ---------------------------------- */
    cur = NotationList;
    while( cur != NULL ) {
        prev = cur;
        cur = cur->Next;
        free( prev->BlackName );
        free( prev->BlackAddress );
        free( prev->WhiteName );
        free( prev->WhiteAddress );
        free( prev );
    }
    NotationList = NULL;
} /* CleanupNotationList */

/* ** Func NotationNum *************************** */
/* Return number of notations in Notation List. */
int NotationNum( void )
{
/* -- Local Data --------------------------------- */
    NOTATIONDATA *cur;
    int ret;
/* -- Operation ---------------------------------- */
    ret = 0;
    cur = NotationList;
    while( cur != NULL ) {
        ret++;
        cur = cur->Next;
    }
    return ret;
} /* NotationNum */


/* ** Func ManageNotation **************************** */
/*
 * Check notation file.
 * If notation size is too big, delete oldest notation.
 * If expire date of a notation has come, expire the notation.
 * return 0 -- success.
 * return 1 -- fail.
 */
int ManageNotation( void )
{
/* -- Local Data --------------------------------- */
    FILE *fp;
    FILE *tmp;
    long pos;
    char buf[257];
    char tmpbuf[257];
    int isNotationFileEnd;
    int ShouldCreateNewNotationFile;
    int i;
    time_t begin;
    time_t end;
    time_t now;
    long ShouldDeleteBytes;
    int ShouldCopyThisNotation;
/* -- Operation ---------------------------------- */
    tmp = tmpfile();
    if( tmp == NULL ) {
        Err( "Cannot open temporary file." );
        return 1;
    }
    fp = fopen( NotationFile, "r" );
    if( fp == NULL ) {
        Err( "Cannot open notation file %s.", NotationFile );
        return 1;
    }
    /* writing ..... */
    while( 1 ) {
        isNotationFileEnd = FALSE;
        ShouldCreateNewNotationFile = TRUE;
        /* Search Next TUITATE_NOTATION magic */
        while( 1 ) {
            pos = ftell( fp );  /* Top of a notation block */
            if( fgets( buf, 256, fp ) == NULL ) {
                isNotationFileEnd = TRUE;
                ShouldCreateNewNotationFile = TRUE;
                break;
            }
            if( !strncmp( buf, MAGIC, strlen( MAGIC ) ) ) {
                break;
            }
        }
        if( isNotationFileEnd ) break;
        if( strstr( buf, STR_SAVE ) == NULL ) { /* not have save propery */
            /* Begin time */
            if( fgets( buf, 256, fp ) == NULL ) {
                /* strange notation file */
                isNotationFileEnd = TRUE;
                ShouldCreateNewNotationFile = FALSE;
                break;
            }
            for( i = 0; isdigit( buf[i] ); i++ ) tmpbuf[i] = buf[i];
            tmpbuf[i] = '\0';
            begin = atol( tmpbuf );
            /* End time */
            if( fgets( buf, 256, fp ) == NULL ) {
                /* strange notation file */
                isNotationFileEnd = TRUE;
                ShouldCreateNewNotationFile = FALSE;
                break;
            }
            for( i = 0; isdigit( buf[i] ); i++ ) tmpbuf[i] = buf[i];
            tmpbuf[i] = '\0';
            end = atol( tmpbuf );
            /* Now time */
            now = time( NULL );
            if( now - end < ExpireDays * 24 * 60 * 60 )
                ShouldCopyThisNotation = TRUE;
            else
                ShouldCopyThisNotation = FALSE;
        } else {                           /* have save propery */
            ShouldCopyThisNotation = TRUE;
        }
        if( ShouldCopyThisNotation ) {
            /* copy this notation part into new notation file */
            fseek( fp, pos, SEEK_SET );
            while( 1 ) {
                if( fgets( buf, 256, fp ) == NULL ) {
                    /* strange notation file */
                    isNotationFileEnd = TRUE;
                    ShouldCreateNewNotationFile = FALSE;
                    Err( "Notation File is strange." );
                    break;
                }
                fputs( buf, tmp );
                if( !strncmp( buf, KEY_END, strlen( KEY_END ) ) )
                    break;
            }
        } else {
            /* skip this notation part */
            while( 1 ) {
                if( fgets( buf, 256, fp ) == NULL ) {
                    /* strange notation file */
                    isNotationFileEnd = TRUE;
                    ShouldCreateNewNotationFile = FALSE;
                    Err( "Notation File is strange." );
                    break;
                }
                if( !strncmp( buf, KEY_END, strlen( KEY_END ) ) )
                    break;
            }
        }
        if( isNotationFileEnd ) break;
    }
    if( !ShouldCreateNewNotationFile ) {
        fclose( tmp );
        fclose( fp );
        return 1;
    } else {
        fclose( fp );
        remove( NotationFile );
        fp = fopen( NotationFile, "w" );
        if( fp == NULL ) {
            Err( "Cannot open notation file %s.", NotationFile );
            fclose( tmp );
            return 1;
        }
        fseek( tmp, 0L, SEEK_SET );
        while( 1 ) {
            if( fgets( buf, 256, tmp ) == NULL )
                break;
            if( fputs( buf, fp ) == EOF ) {
                fclose( tmp );
                fclose( fp );
                Err( "Notation write error." );
                return 1;
            }
        }
        fclose( tmp );
        pos = ftell( fp );
        fclose( fp );
    }
    /* Next notation file size check */
    ShouldDeleteBytes = pos - NOTATIONFILESIZE;
    if( ShouldDeleteBytes > 0 ) {
        tmp = tmpfile();
        if( tmp == NULL ) {
            Err( "Cannot open temporary file." );
            return 1;
        }
        fp = fopen( NotationFile, "r" );
        if( fp == NULL ) {
            fclose( tmp );
            Err( "Cannot open notation file." );
            return 1;
        }
        /* Skip notaions from head of notation file */
        while( 1 ) {
            if( fgets( buf, 256, fp ) == NULL ) {
                fclose( tmp );
                fclose( fp );
                Err( "Strange notation file." );
                return 1;
            }
            ShouldDeleteBytes -= strlen( buf );
            if( !strncmp( buf, KEY_END, strlen( KEY_END ) ) ) {
                if( ShouldDeleteBytes < 0 ) break;
            }
        }
        /* copy notations should be remained */
        while( 1 ) {
            if( fgets( buf, 256, fp ) == NULL ) break;
            fputs( buf, tmp );
        }
        fclose( fp );
        /* create new notation file */
        fp = fopen( NotationFile, "w" );
        if( fp == NULL ) {
            fclose( tmp );
            Err( "Cannot open notation file." );
            return 1;
        }
        fseek( tmp, 0L, SEEK_SET );
        while( 1 ) {
            if( fgets( buf, 256, tmp ) == NULL )
                break;
            if( fputs( buf, fp ) == EOF ) {
                fclose( tmp );
                fclose( fp );
                Err( "Notation write error." );
                return 1;
            }
        }
        fclose( tmp );
        fclose( fp );
    }
    return 0;
} /* ManageNotation */

/* ** Func DeleteNotation **************************** */
/*
 * Delete specified notation from notation file.
 * return 0 -- success.
 * return 1 -- fail.
 */
int DeleteNotation( int NotationID )
{
/* -- Local Data --------------------------------- */
    FILE *fp;
    FILE *tmp;
    char buf[257];
    int isNotationFileEnd;
    int ShouldCreateNewNotationFile;
    long ShouldWrite;
    int id;
/* -- Operation ---------------------------------- */
    fp = fopen( NotationFile, "r" );
    if( fp == NULL ) {
        Err( "Cannot open notation file %s.", NotationFile );
        return 1;
    }
    /* check if the notation exist or not */
    ShouldCreateNewNotationFile = FALSE;
    isNotationFileEnd = FALSE;
    while( 1 ) {
        /* Search Next TUITATE_NOTATION magic */
        while( 1 ) {
            if( fgets( buf, 256, fp ) == NULL ) {
                isNotationFileEnd = TRUE;
                break;
            }
            if( !strncmp( buf, MAGIC, strlen( MAGIC ) ) ) {
                break;
            }
        }
        if( isNotationFileEnd ) break;
        id = GetNumberFromString( buf );
        if( id == NotationID ) { /* found! */
            ShouldCreateNewNotationFile = TRUE;
            break;
        }
    }
    if( !ShouldCreateNewNotationFile ) {
        fclose( fp );
        return 1;
    } else {
        tmp = tmpfile();
        if( tmp == NULL ) {
            Err( "Cannot open temporary file." );
            fclose( fp );
            return 1;
        }
        fseek( fp, 0, SEEK_SET );
        ShouldWrite = TRUE;
        while( 1 ) {
            if( fgets( buf, 256, fp ) == NULL ) /* notation end */
                break;
            if( !strncmp( buf, MAGIC, strlen( MAGIC ) ) ) {
                id = GetNumberFromString( buf );
                if( id == NotationID )
                    ShouldWrite = FALSE;
                else if( !ShouldWrite )
                    ShouldWrite = TRUE;
            }
            if( ShouldWrite )
                fputs( buf, tmp );
        }
        fclose( fp );
        remove( NotationFile );
        fp = fopen( NotationFile, "w" );
        if( fp == NULL ) {
            fclose( tmp );
            Err( "Cannot open notation file." );
            return 1;
        }
        fseek( tmp, 0, SEEK_SET );
        while( fgets( buf, 256, tmp ) != NULL ) {
            fputs( buf, fp );
        }
        fclose( tmp );
        fclose( fp );
    }
    UpdateNotationList();
    return 0;
} /* DeleteNotation */

/* ** Func AddSavePropertyToNotation **************************** */
/*
 * Add save property to specified notation,
 * and write it into the notation file.
 * return 0 -- success.
 * return 1 -- fail.
 */
int AddSavePropertyToNotation( int NotationID )
{
/* -- Local Data --------------------------------- */
    FILE *fp;
    FILE *tmp;
    char buf[257];
    int isNotationFileEnd;
    int ShouldCreateNewNotationFile;
    int id;
/* -- Operation ---------------------------------- */
    fp = fopen( NotationFile, "r" );
    if( fp == NULL ) {
        Err( "Cannot open notation file %s.", NotationFile );
        return 1;
    }
    /* count notation number and check if the notation exist or not */
    ShouldCreateNewNotationFile = FALSE;
    isNotationFileEnd = FALSE;
    while( 1 ) {
        /* Search Next TUITATE_NOTATION magic */
        while( 1 ) {
            if( fgets( buf, 256, fp ) == NULL ) {
                isNotationFileEnd = TRUE;
                break;
            }
            if( !strncmp( buf, MAGIC, strlen( MAGIC ) ) ) {
                break;
            }
        }
        if( isNotationFileEnd ) break;
        id = GetNumberFromString( buf );
        if( id == NotationID ) { /* found! */
            ShouldCreateNewNotationFile = TRUE;
            break;
        }
    }
    if( !ShouldCreateNewNotationFile ) {
        fclose( fp );
        return 1;
    } else {
        tmp = tmpfile();
        if( tmp == NULL ) {
            Err( "Cannot open temporary file." );
            fclose( fp );
            return 1;
        }
        fseek( fp, 0, SEEK_SET );
        while( 1 ) {
            if( fgets( buf, 256, fp ) == NULL ) /* notation end */
                break;
            if( !strncmp( buf, MAGIC, strlen( MAGIC ) ) ) {
                id = GetNumberFromString( buf );
                if( id == NotationID ) {
                    if( strstr( buf, STR_SAVE ) == NULL )
                        sprintf( buf, "%s %d %s\n", MAGIC, id, STR_SAVE );
                }
            }
            fputs( buf, tmp );
        }
        fclose( fp );
        remove( NotationFile );
        fp = fopen( NotationFile, "w" );
        if( fp == NULL ) {
            fclose( tmp );
            Err( "Cannot open notation file." );
            return 1;
        }
        fseek( tmp, 0, SEEK_SET );
        while( fgets( buf, 256, tmp ) != NULL ) {
            fputs( buf, fp );
        }
        fclose( tmp );
        fclose( fp );
    }
    UpdateNotationList();
    return 0;
} /* AddSavePropertyToNotation */

/* ** Func RemoveSavePropertyFromNotation **************************** */
/*
 * Remove save property from specified notation,
 * and write it into the notation file.
 * return 0 -- success.
 * return 1 -- fail.
 */
int RemoveSavePropertyFromNotation( int NotationID )
{
/* -- Local Data --------------------------------- */
    FILE *fp;
    FILE *tmp;
    char buf[257];
    int isNotationFileEnd;
    int ShouldCreateNewNotationFile;
    int id;
/* -- Operation ---------------------------------- */
    fp = fopen( NotationFile, "r" );
    if( fp == NULL ) {
        Err( "Cannot open notation file %s.", NotationFile );
        return 1;
    }
    /* count notation number and check if the notation exist or not */
    ShouldCreateNewNotationFile = FALSE;
    isNotationFileEnd = FALSE;
    while( 1 ) {
        /* Search Next TUITATE_NOTATION magic */
        while( 1 ) {
            if( fgets( buf, 256, fp ) == NULL ) {
                isNotationFileEnd = TRUE;
                ShouldCreateNewNotationFile = FALSE;
                break;
            }
            if( !strncmp( buf, MAGIC, strlen( MAGIC ) ) ) {
                break;
            }
        }
        if( isNotationFileEnd ) break;
        id = GetNumberFromString( buf );
        if( id == NotationID ) { /* found! */
            if( strstr( buf, STR_SAVE ) != NULL )
                ShouldCreateNewNotationFile = TRUE;
            else
                ShouldCreateNewNotationFile = FALSE;
            break;
        }
    }
    if( !ShouldCreateNewNotationFile ) {
        fclose( fp );
        return 1;
    } else {
        tmp = tmpfile();
        if( tmp == NULL ) {
            Err( "Cannot open temporary file." );
            fclose( fp );
            return 1;
        }
        fseek( fp, 0, SEEK_SET );
        while( 1 ) {
            if( fgets( buf, 256, fp ) == NULL ) /* notation end */
                break;
            if( !strncmp( buf, MAGIC, strlen( MAGIC ) ) ) {
                id = GetNumberFromString( buf );
                if( id == NotationID )
                    sprintf( buf, "%s %d\n", MAGIC, id );
            }
            fputs( buf, tmp );
        }
        fclose( fp );
        remove( NotationFile );
        fp = fopen( NotationFile, "w" );
        if( fp == NULL ) {
            fclose( tmp );
            Err( "Cannot open notation file." );
            return 1;
        }
        fseek( tmp, 0, SEEK_SET );
        while( fgets( buf, 256, tmp ) != NULL ) {
            fputs( buf, fp );
        }
        fclose( tmp );
        fclose( fp );
    }
    UpdateNotationList();
    return 0;
} /* RemoveSavePropertyFromNotation */

/* ** Func GenerateNotationID **************************** */
/*
 * Decide New Notation ID number.
 * Return new ID.
 * return 0 -- fail.
 */
int GenerateNotationID( void )
{
/* -- Local Data --------------------------------- */
    int id;
    int start;
    NOTATIONDATA *p;
    int TryNext;
/* -- Operation ---------------------------------- */
    start = id = NewestNotationID();
    if( id == 0 ) {
        if( NotationList != NULL ) return 0;
        else return 1;
    }
    while( 1 ) {
        id++;
        if( id > MAXNOTATIONNUM ) id = 1;
        if( id == start ) return 1; /* id search have cycled */
        TryNext = FALSE;
        for( p = NotationList; p != NULL; p = p->Next ) {
            if( p->ID == id ) {
                TryNext = TRUE;
                break;
            }
        }
        if( !TryNext ) {
            break;
        }
    }
    return id;
} /* GenerateNotationID */

/* ** Func NewestNotationID **************************** */
/*
 * Return ID of the newest notation.
 * Return the ID.
 * return 0 -- fail (not found).
 */
int NewestNotationID( void )
{
/* -- Local Data --------------------------------- */
    int id;
    NOTATIONDATA *p;
    time_t NewestTime;
/* -- Operation ---------------------------------- */
    id = 0;
    NewestTime = 0;
    for( p = NotationList; p != NULL; p = p->Next ) {
        if( p->End >= NewestTime ) {
            NewestTime = p->End;
            id = p->ID;
        }
    }
    return id;
} /* NewestNotationID */


/* ** Func GetNumberFromString **************************** */
/*
 * Return number included in the given string as decimal.
 * return 0 -- fail (not found).
 * It is assumed that only one plus number is included in it,
 * and the length of the string is less than 256.
 */
int GetNumberFromString( const char *str )
{
/* -- Local Data --------------------------------- */
    const char *p;
    char buf[256];
    int i;
/* -- Operation ---------------------------------- */
    for( p = str; !isdigit( *p ) ; p++ )
        if( *p == '\0' ) return 0;
    for( i = 0; isdigit( p[i] ); i++ ) buf[i] = p[i];
    buf[i] = '\0';
    return atol( buf );
} /* GetNumberFromString */




