/* ** receivemessage.c *************************** */
/*
 * Impliment routines to receive messages
 * from clientes in TUITATE PROTOCOL.
 */
/* *********************************************** */

/*
 * PACKET : PARAM PARAM ...
 *			Start with head of line, and End with '\n'
 *      PARAMs are separated with chars which match
 *      "isspace" function.
 *			One PACKET must not be longer than 256 byte.
 */

/* -- Include File ------------------------------- */
#include <config.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#include <sys/uio.h>
#include "tuitated.h"
#include "struct.h"
#include "protocol.h"

/* -- Macro Definition --------------------------- */
#define HASHSIZE 200

/* -- Protype Declerations  ---------------------- */
int ParamNum( char *line );
int GetParam( char *line, int nth, char *buf );
int GetParamNotConverted( char *line, int nth, char *buf );
CLIENT_MSGKEY ReadMessage( CLIENTDATA *client, CLIENT_MESSAGE *msg, int *ShouldSelect );
int StrToHashKey( char *str );
CLIENT_MSGKEY ReceiveMode( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveHelp( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveQuit( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveView( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveNotation( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveManagenotation( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveGet( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveGame( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveChallengerlist( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveChallenge( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveAccept( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveRefuse( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveMove( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveDrop( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveGiveup( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveCheat( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveGamelist( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveViewgame( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveNotationlist( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveViewnotation( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveDeletenotation( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveSavenotation( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveUnsavenotation( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveEmptyLine( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveNotYetOnePacket( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveEmergency( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
CLIENT_MSGKEY ReceiveError( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
int mystrtoi( char *str );


/* -- Global Data -------------------------------- */
CLIENT_MSGKEY (*(ReceiveMessageHandlers[MSGKEY_NUM]))( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg ) = {
    ReceiveMode,
    ReceiveHelp,
    ReceiveQuit,
    ReceiveGame,
    ReceiveView,
    ReceiveNotation,
    ReceiveManagenotation,
    ReceiveGet,
    ReceiveChallengerlist,
    ReceiveChallenge,
    ReceiveAccept,
    ReceiveRefuse,
    ReceiveMove,
    ReceiveDrop,
    ReceiveGiveup,
    ReceiveCheat,
    ReceiveGamelist,
    ReceiveViewgame,
    ReceiveNotationlist,
    ReceiveViewnotation,
    ReceiveDeletenotation,
    ReceiveSavenotation,
    ReceiveUnsavenotation,
    ReceiveEmptyLine,
    ReceiveNotYetOnePacket,
    ReceiveEmergency,
    ReceiveError,
};
typedef struct _MSG_HANDLER {
    char *keyword;
    CLIENT_MSGKEY (* handler)( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg );
} MSG_HANDLER;
MSG_HANDLER ReceiveMessageHashTable[HASHSIZE];


/* ** Func InitReceiveHandlers ****************************** */
/*
 * Initialize Hash Table of Handlers which receive messages.
 */
void InitReceiveHandlers( void )
{
/* -- Local Data --------------------------------- */
    int i;
/* -- Operation ---------------------------------- */
    for( i = 0; i < HASHSIZE; i++ ) {
        ReceiveMessageHashTable[i].keyword = NULL;
        ReceiveMessageHashTable[i].handler = NULL;
    }
    for( i = 0; i < MSGKEY_NUM; i++ ) {
        int key;
        key = StrToHashKey( ClientMessageStr[i] );
        while( ReceiveMessageHashTable[key].handler != NULL ) {
            key++;
            if( key >= HASHSIZE ) key = 0;
        }
        ReceiveMessageHashTable[key].keyword = ClientMessageStr[i];
        ReceiveMessageHashTable[key].handler = ReceiveMessageHandlers[i];
    }
} /* InitReceiveHandlers */
	
/* ** Func StrToHashKey ****************************** */
/*
 * Sum up all chars in given string and get mod of it.
 */
int StrToHashKey( char *str )
{
/* -- Local Data --------------------------------- */
    int n;
    char *p;
/* -- Operation ---------------------------------- */
    for( p = str, n = 0; *p != '\0'; n += *p++ );
    return( n % HASHSIZE );
} /* StrToHashKey */

/* ** Func ParamNum ****************************** */
/*
 * Return number of params in given string line.
 */
int ParamNum( char *line )
{
/* -- Local Data --------------------------------- */
    int isSpace; /* current examining char is space */
    char ret;
    char *p;
/* -- Operation ---------------------------------- */
    isSpace = TRUE;
    ret = FALSE;
    p = line;
    while( 1 ) {
        if( *p == '\0' ) {
            break;
        } else if( isspace( *p ) ) {
            if( !isSpace ) {
                isSpace = TRUE;
            }
        } else {
            if( isSpace ) {
                ret++;
                isSpace = FALSE;
            }
        }
        p++;
    }
    return ret;
} /* ParamNum */
	
/* ** Func GetParam ****************************** */
/*
 * Get One PARAM of given line( copy to given buf ),
 * which is nth ( 0, 1, ... ) in the line.
 * Return 0 .. OK.
 * Return 1 .. Error.
 */
int GetParam( char *line, int nth, char *buf )
{
/* -- Local Data --------------------------------- */
    int nthParam = 0;
    int isSpace = TRUE;
    int i = 0;
    char *p;
/* -- Operation ---------------------------------- */
    p = line;
    if( nth < 0 ) return 1;
    /* skip until nth PARAM appear */
    while( 1 ) {
        if( *p == '\0' ) {
            return 1;
        } else if( isspace( *p ) ) {
            if( !isSpace ) {
                isSpace = TRUE;
                nthParam++;
            }
        } else {
            if( isSpace ) {
                if( nth == nthParam ) {
                    break;
                } else {
                    isSpace = FALSE;
                }
            }
        }
        p++;
    }
    /* copy nth PARAM to given buffer */
    for( i = 0; ; i++ ) {
        buf[i] = toupper( p[i] );
        if( buf[i] == '\0') break;
        if( isspace( buf[i] ) ) {
            buf[i] = '\0';
            break;
        }
    }
    return 0;
} /* GetParam */
	
/* ** Func GetParamNotConverted ******************* */
/*
 * Get One PARAM of given line( copy to given buf ),
 * which is nth ( 0, 1, ... ) in the line.
 * This func is different from the func GetParam
 * on the point that this func do not convert chars
 * into upper case.
 * Return 0 .. OK.
 * Return 1 .. Error.
 */
int GetParamNotConverted( char *line, int nth, char *buf )
{
/* -- Local Data --------------------------------- */
    int nthParam = 0;
    int isSpace = TRUE;
    int i = 0;
    char *p;
/* -- Operation ---------------------------------- */
    p = line;
    if( nth < 0 ) return 1;
    /* skip until nth PARAM appear */
    while( 1 ) {
        if( *p == '\0' ) {
            return 1;
        } else if( isspace( *p ) ) {
            if( !isSpace ) {
                isSpace = TRUE;
                nthParam++;
            }
        } else {
            if( isSpace ) {
                if( nth == nthParam ) {
                    break;
                } else {
                    isSpace = FALSE;
                }
            }
        }
        p++;
    }
    /* copy nth PARAM to given buffer */
    for( i = 0; ; i++ ) {
        buf[i] = p[i];
        if( buf[i] == '\0') break;
        if( isspace( buf[i] ) ) {
            buf[i] = '\0';
            break;
        }
    }
    return 0;
} /* GetParamNotConverted */


/* ** Func ReturnCodeExist **************************** */
/*
 * Check if the given string contains return code or not.
 * return code := (0xd0xa or 0xa)
 * return 1 -- contaion return code.
 * return 0 -- NOT contaion return code.
 */
int ReturnCodeExist( char *str )
{
    char *p = str;
    while( *p++ ) {
        if( *p == 0xd && *(p+1) == 0xa ) {
            return 1;
        }
        if( *p == 0xa ) {
            return 1;
        }
    }
    return 0;
} /* ReturnCodeExist */

/* ** Func ParseMessage ****************************** */
/*
 * Parse message in given line.
 */
CLIENT_MSGKEY ParseMessage( char *LineBuf, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
    int ParamNumber;
    int OrgKey;
    int key;
    char buf[257];
/* -- Operation ---------------------------------- */
    if( strlen( LineBuf ) == 0 ) {/* Special handling for Empty Line message */
        return EMPTYLINE;
    }

    ParamNumber = ParamNum( LineBuf );

    /* Get first PARAM into buf */
    if( GetParam( LineBuf, 0, buf ) ) {
        Err( "ReadMessage: fail to get first PARAM." );
        return ERROR;
    }
    /* Parse Key */
    OrgKey = key = StrToHashKey( buf );
    do {
        if( ReceiveMessageHashTable[key].keyword != NULL ) {
            if( !strcmp( buf, ReceiveMessageHashTable[key].keyword ) ) {
                return ReceiveMessageHashTable[key].handler
                    ( LineBuf, ParamNumber, msg );
            }
        }
        key++;
        if( key >= HASHSIZE ) key = 0;
    } while( key != OrgKey ); /* Search hash table */
    return ERROR;
} /* ParseMessage */

/* ** Func ReadMessage **************************** */
/*
 * Read one packet( one line ), and parse it.
 * Return TYPE of Packet,
 * Return TYPE_ERROR if error occurs.
 * Return MSGKEY_NUM if empty line received.
 */
CLIENT_MSGKEY ReadMessage( CLIENTDATA *client, CLIENT_MESSAGE *msg, int *ShouldSelect )
{
/* -- Local Data --------------------------------- */
    int ReadRet;
    int CharNum;
    char *p;
    char LineBuf[256];
    char buf[CLIENT_BUFSIZE];
    CLIENT_MSGKEY ret;
/* -- Operation ---------------------------------- */
    /* Check if the buffer contain whole one message. */
    if( !ReturnCodeExist( client->ReceiveMsgBuf ) ) {
        /* No return code. Then read data from client. */
        CharNum = strlen( client->ReceiveMsgBuf );
        ReadRet = read( client->soc,
                        client->ReceiveMsgBuf + CharNum,
                        sizeof( client->ReceiveMsgBuf ) - CharNum - 1 );
        if( ReadRet == 0 ) { /* EOF ( socket closed by client ) */
            Err( "read. -- EOF" );
            *ShouldSelect = TRUE;
            return EMERGENCY;
        } else if( ReadRet < 0 ) {
            if( errno == EAGAIN ) { /* no data is in socket */
                *ShouldSelect = TRUE;
                return NOTYETONEPACKET;
            } else {
                Err( "read." );
                *ShouldSelect = TRUE;
                return EMERGENCY;
            }
        } else {  /* Data is gotten */
            p = client->ReceiveMsgBuf + CharNum;
            CharNum += ReadRet;
            client->ReceiveMsgBuf[CharNum] = '\0';
            AccessLog( "from client %d '%s'", client->soc, p );
            /* check for control codes in message */
            for( ; *p; p++ ) {
                if( !(isprint(*p) || *p == '\t' || *p == 0xd || *p == 0xa) ) {
                    Err("Control Code in Message from client %d", client->soc);
                    *ShouldSelect = TRUE;
                    return EMERGENCY;
                }
            }
            /* Check if the buffer contain whole one message. */
            if( !ReturnCodeExist( client->ReceiveMsgBuf ) ) {
                if( strlen( client->ReceiveMsgBuf ) > 256 ) {
                    Err( "Illegal message from client %d.", client->soc );
                    *ShouldSelect = TRUE;
                    return EMERGENCY;
                }
                return NOTYETONEPACKET;
            }
        }
    }
    /* More than one message is exist in client->ReceiveMsgBuf... */
    for( p = client->ReceiveMsgBuf; *p; p++ ) {
        if( ((*p == 0xd) && (*(p+1) == 0xa)) || ( *p == 0xa ) ) {
            /* Found!  Let's parse it! */
            if( *p == 0xd ) {
                strcpy( buf, p+2 );
            } else {
                strcpy( buf, p+1 );
            }
            *p = '\0';
            if( strlen( client->ReceiveMsgBuf ) > 256 ) {
                Err( "Illegal message from client %d.", client->soc );
                *ShouldSelect = TRUE;
                return EMERGENCY;
            }
            strcpy( LineBuf, client->ReceiveMsgBuf );
            strcpy( client->ReceiveMsgBuf, buf );
            ret = ParseMessage( LineBuf, msg );
            if( ReturnCodeExist( client->ReceiveMsgBuf ) ) {
                switch( ret ) {
                  case QUIT:
                  case EMERGENCY:
                    *ShouldSelect = TRUE; break;
                  default:
                    *ShouldSelect = FALSE; break;
                }
            } else {
                *ShouldSelect = TRUE;
            }
            return ret;
        }
    }
    return ERROR; /* why reach here? */
} /* ReadMessage */



/* ** Func ReceiveMode ****************************** */
/*
 * Parse given line as MODE message.
 */
CLIENT_MSGKEY ReceiveMode( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
    msg->key = MODE;
    return MODE;
} /* ReceiveMode */

/* ** Func ReceiveHelp ****************************** */
/*
 * Parse given line as HELP message.
 */
CLIENT_MSGKEY ReceiveHelp( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
    msg->key = HELP;
    return HELP;
} /* ReceiveHelp */

/* ** Func ReceiveQuit ****************************** */
/*
 * Parse given line as QUIT message.
 */
CLIENT_MSGKEY ReceiveQuit( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
    msg->key = QUIT;
    return QUIT;
} /* ReceiveQuit */

/* ** Func ReceiveView ****************************** */
/*
 * Parse given line as VIEW message.
 */
CLIENT_MSGKEY ReceiveView( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
    msg->key = VIEW;
    return VIEW;
} /* ReceiveView */

/* ** Func ReceiveNotation ****************************** */
/*
 * Parse given line as NOTATION message.
 */
CLIENT_MSGKEY ReceiveNotation( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
    msg->key = NOTATION;
    return NOTATION;
} /* ReceiveNotation */

/* ** Func ReceiveManagenotation ****************************** */
/*
 * Parse given line as MANAGENOTATION message.
 */
CLIENT_MSGKEY ReceiveManagenotation( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
    msg->key = MANAGENOTATION;
    return MANAGENOTATION;
} /* ReceiveManagenotation */

/* ** Func ReceiveGet ****************************** */
/*
 * Parse given line as GET message.
 */
CLIENT_MSGKEY ReceiveGet( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
    char buf[256];
/* -- Operation ---------------------------------- */
    if( GetParamNotConverted( LineBuf, 1, buf ) ) return ERROR;
    msg->name = malloc( strlen( buf ) + 1 );
    if( msg->name == NULL ) return ERROR;
    strcpy( msg->name, buf );
    if( GetParamNotConverted( LineBuf, 2, buf ) ) {
        free( msg->name );
        return ERROR;
    } else if( strncmp( buf, GET_END, strlen( GET_END ) ) ) {
        free( msg->name );
        return ERROR;
    }
    msg->key = GET;
    return GET;
} /* ReceiveGet */

/* ** Func ReceiveGame ****************************** */
/*
 * Parse given line as GAME message.
 */
CLIENT_MSGKEY ReceiveGame( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
    char buf[256];
/* -- Operation ---------------------------------- */
    if( !( ParamNumber == 1 || ParamNumber == 2 || ParamNumber == 3 ) ) {
        return ERROR;
    }
    if( ParamNumber > 1 ) {
        if( GetParamNotConverted( LineBuf, 1, buf ) ) return ERROR;
        msg->name = malloc( strlen( buf ) + 1 );
        if( msg->name == NULL ) return ERROR;
        strcpy( msg->name, buf );
        if( ParamNumber > 2 ) {
            if( GetParamNotConverted( LineBuf, 2, buf ) ) return ERROR;
            msg->mailaddress = malloc( strlen( buf ) + 1 );
            if( msg->mailaddress == NULL ) return ERROR;
            strcpy( msg->mailaddress, buf );
        } else {
            msg->mailaddress = malloc( strlen( "Unknown" ) + 1 );
            if( msg->mailaddress == NULL ) return ERROR;
            strcpy( msg->mailaddress, "Unknown" );
        }
    } else {
        msg->name = malloc( strlen( "Unknown" ) + 1 );
        if( msg->name == NULL ) return ERROR;
        strcpy( msg->name, "Unknown" );
        msg->mailaddress = malloc( strlen( "Unknown" ) + 1 );
        if( msg->mailaddress == NULL ) return ERROR;
        strcpy( msg->mailaddress, "Unknown" );
    }
    msg->key = GAME;
    return GAME;
} /* ReceiveGame */

/* ** Func ReceiveChallengerlist ****************************** */
/*
 * Parse given line as CHALLENGERLIST message.
 */
CLIENT_MSGKEY ReceiveChallengerlist( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
    msg->key = CHALLENGERLIST;
    return CHALLENGERLIST;
} /* ReceiveChallengerlist */

/* ** Func ReceiveChallenge ****************************** */
/*
 * Parse given line as CHALLENGE message.
 */
CLIENT_MSGKEY ReceiveChallenge( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
    char buf[256];
/* -- Operation ---------------------------------- */
    if( !(ParamNumber == 2) ) return ERROR;
    if( GetParam( LineBuf, 1, buf ) ) return ERROR;
    msg->num = mystrtoi( buf );
    msg->key = CHALLENGE;
    return CHALLENGE;
} /* ReceiveChallenge */

/* ** Func ReceiveAccept ****************************** */
/*
 * Parse given line as ACCEPT message.
 */
CLIENT_MSGKEY ReceiveAccept( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
    msg->key = ACCEPT;
    return ACCEPT;
} /* ReceiveAccept */

/* ** Func ReceiveRefuse ****************************** */
/*
 * Parse given line as REFUSE message.
 */
CLIENT_MSGKEY ReceiveRefuse( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
    msg->key = REFUSE;
    return REFUSE;
} /* ReceiveRefuse */

/* ** Func ReceiveMove ****************************** */
/*
 * Parse given line as MOVE message.
 */
CLIENT_MSGKEY ReceiveMove( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
    char buf[256];
/* -- Operation ---------------------------------- */
    if( !(ParamNumber == 5 || ParamNumber == 6) ) return ERROR;
    msg->key = MOVE;
    /* get sx */
    if( GetParam( LineBuf, 1, buf ) ) return ERROR;
    if( strlen( buf ) != 1 )  return ERROR;
    if( buf[0] < '1' || '9' < buf[0]  ) return ERROR;
    msg->hand.move.sx = buf[0] - '1';
    /* get sy */
    if( GetParam( LineBuf, 2, buf ) ) return ERROR;
    if( strlen( buf ) != 1 )  return ERROR;
    if( buf[0] < '1' || '9' < buf[0]  ) return ERROR;
    msg->hand.move.sy = buf[0] - '1';
    /* get dx */
    if( GetParam( LineBuf, 3, buf ) ) return ERROR;
    if( strlen( buf ) != 1 )  return ERROR;
    if( buf[0] < '1' || '9' < buf[0]  ) return ERROR;
    msg->hand.move.dx = buf[0] - '1';
    /* get dy */
    if( GetParam( LineBuf, 4, buf ) ) return ERROR;
    if( strlen( buf ) != 1 )  return ERROR;
    if( buf[0] < '1' || '9' < buf[0]  ) return ERROR;
    msg->hand.move.dy = buf[0] - '1';
    /* get isnaru */
    if( ParamNumber == 6 ) {
        if( GetParam( LineBuf, 5, buf ) ) return ERROR;
        if( !strcmp( buf, STR_PROMOTE ) ) {
            msg->hand.move.willPromote = TRUE;
        } else if( !strcmp( buf, STR_NOTPROMOTE ) ) {
            msg->hand.move.willPromote = FALSE;
        } else {
            return ERROR;
        }
    } else {
        msg->hand.move.willPromote = FALSE;
    }
    return MOVE;
} /* ReceiveMove */

/* ** Func ReceiveDrop ****************************** */
/*
 * Parse given line as DROP message.
 */
CLIENT_MSGKEY ReceiveDrop( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
    int i;
    char buf[256];
/* -- Operation ---------------------------------- */
    if( ParamNumber != 4 ) return ERROR;
    msg->key = DROP;
    /* get dx */
    if( GetParam( LineBuf, 1, buf ) ) return ERROR;
    if( strlen( buf ) != 1 )  return ERROR;
    if( buf[0] < '1' || '9' < buf[0]  ) return ERROR;
    msg->hand.drop.dx = buf[0] - '1';
    /* get dy */
    if( GetParam( LineBuf, 2, buf ) ) return ERROR;
    if( strlen( buf ) != 1 )  return ERROR;
    if( buf[0] < '1' || '9' < buf[0]  ) return ERROR;
    msg->hand.drop.dy = buf[0] - '1';
    /* get piecetype */
    if( GetParam( LineBuf, 3, buf ) ) return ERROR;
    for( i = NONE; i < PIECE_NUM; i++ ) {
        if( !strcmp( buf, PieceStr[i] ) ) {
            msg->hand.drop.piece = i;
            break;
        }
    }
    return DROP;
} /* ReceiveDrop */

/* ** Func ReceiveGiveup ****************************** */
/*
 * Parse given line as GIVEUP message.
 */
CLIENT_MSGKEY ReceiveGiveup( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
    msg->key = GIVEUP;
    return GIVEUP;
} /* ReceiveGiveup */

/* ** Func ReceiveCheat ****************************** */
/*
 * Parse given line as CHEAT message.
 */
CLIENT_MSGKEY ReceiveCheat( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
    char buf[256];
/* -- Operation ---------------------------------- */
    if( ParamNumber != 3 ) return ERROR;
    msg->key = CHEAT;
    if( GetParam( LineBuf, 1, buf ) ) return ERROR;
    msg->hand.move.sx = mystrtoi( buf );
    if( GetParam( LineBuf, 2, buf ) ) return ERROR;
    msg->hand.move.sy = mystrtoi( buf );
    return CHEAT;
} /* ReceiveCheat */

/* ** Func ReceiveGamelist ****************************** */
/*
 * Parse given line as GAMELIST message.
 */
CLIENT_MSGKEY ReceiveGamelist( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
    msg->key = GAMELIST;
    return GAMELIST;
} /* ReceiveGamelist */

/* ** Func ReceiveViewgame ****************************** */
/*
 * Parse given line as VIEWGAME message.
 */
CLIENT_MSGKEY ReceiveViewgame( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
    char buf[256];
/* -- Operation ---------------------------------- */
    msg->key = VIEWGAME;
    if( GetParam( LineBuf, 1, buf ) ) return ERROR;
    msg->num = mystrtoi( buf );
    return VIEWGAME;
} /* ReceiveViewgame */

/* ** Func ReceiveNotationlist ****************************** */
/*
 * Parse given line as NOTATIONLIST message.
 */
CLIENT_MSGKEY ReceiveNotationlist( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
    msg->key = NOTATIONLIST;
    return NOTATIONLIST;
} /* ReceiveNotationlist */

/* ** Func ReceiveViewnotation ****************************** */
/*
 * Parse given line as VIEWNOTATION message.
 */
CLIENT_MSGKEY ReceiveViewnotation( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
    char buf[256];
/* -- Operation ---------------------------------- */
    msg->key = VIEWNOTATION;
    if( GetParam( LineBuf, 1, buf ) ) return ERROR;
    msg->num = mystrtoi( buf );
    return VIEWNOTATION;
} /* ReceiveViewnotation */

/* ** Func ReceiveDeletenotation ****************************** */
/*
 * Parse given line as DELETENOTATION message.
 */
CLIENT_MSGKEY ReceiveDeletenotation( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
    char buf[256];
/* -- Operation ---------------------------------- */
    msg->key = DELETENOTATION;
    if( GetParam( LineBuf, 1, buf ) ) return ERROR;
    msg->num = mystrtoi( buf );
    return DELETENOTATION;
} /* ReceiveDeletenotation */

/* ** Func ReceiveSavenotation ****************************** */
/*
 * Parse given line as SAVENOTATION message.
 */
CLIENT_MSGKEY ReceiveSavenotation( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
    char buf[256];
/* -- Operation ---------------------------------- */
    msg->key = SAVENOTATION;
    if( GetParam( LineBuf, 1, buf ) ) return ERROR;
    msg->num = mystrtoi( buf );
    return SAVENOTATION;
} /* ReceiveSavenotation */

/* ** Func ReceiveUnsavenotation ****************************** */
/*
 * Parse given line as UNSAVENOTATION message.
 */
CLIENT_MSGKEY ReceiveUnsavenotation( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
    char buf[256];
/* -- Operation ---------------------------------- */
    msg->key = UNSAVENOTATION;
    if( GetParam( LineBuf, 1, buf ) ) return ERROR;
    msg->num = mystrtoi( buf );
    return UNSAVENOTATION;
} /* ReceiveUnsavenotation */

/* ** Func ReceiveEmergency ****************************** */
/*
 * Dummy Function.
 */
CLIENT_MSGKEY ReceiveEmergency( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
    return EMERGENCY;
} /* ReceiveEmergency */

/* ** Func ReceiveEmptyLine ****************************** */
/*
 * Empty Line (only return code) is receivecd.
 */
CLIENT_MSGKEY ReceiveEmptyLine( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
    return EMPTYLINE;
} /* ReceiveEmptyLine */


/* ** Func ReceiveNotYetOnePacket ****************************** */
/*
 * sent message is not enough to parse...
 */
CLIENT_MSGKEY ReceiveNotYetOnePacket( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
    return NOTYETONEPACKET;
} /* ReceiveNotYetOnePacket */


/* ** Func ReceiveError ****************************** */
/*
 * Dummy Function.
 */
CLIENT_MSGKEY ReceiveError( char *LineBuf, int ParamNumber, CLIENT_MESSAGE *msg )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
    return ERROR;
} /* ReceiveError */


/* ** Func mystrtoi ****************************** */
/*
 * Convert string to integer.
 * Illegal input is not concerned.
 */
int mystrtoi( char *str )
{
/* -- Local Data --------------------------------- */
    int ret;
    int len;
    int base;
    int i;
    char *p;
/* -- Operation ---------------------------------- */
    ret = 0;
    for( p = str; *p != '\0'; p++ ) {
        len = strlen( p );
        for( i = 1, base = 1; i < len; i++ ) base *= 10;
        ret += base * ( *p - '0' );
    }
    return ret;
} /* mystrtoi */

