/* ** sendmessage.c ****************************** */
/* Routines to send message from Server. */
/* *********************************************** */

/* -- Include File ------------------------------- */
#include <config.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "tuitated.h"
#include "clientlist.h"
#include "protocol.h"
#include "struct.h"
#include "game.h"
#include "notation.h"
#include "handlist.h"

/* -- Global Data -------------------------------- */
extern char ExportDir[];
extern char NotationFile[];


/* -- Prototype Declerations  -------------------- */
int SendStr( CLIENTDATA *p, const char *str );
int SendData( CLIENTDATA *p, const char *data, int datalen );
int SendHelp( CLIENTDATA *p );
int SendMode( CLIENTDATA *p );
int SendChallengerlist( CLIENTDATA *p );
int SendNotExist( CLIENTDATA * p );
int SendBusy( CLIENTDATA * p );
int SendRefused( CLIENTDATA * p );
int SendChallenged( CLIENTDATA *p, CLIENTDATA *applier );
int SendStart( CLIENTDATA * p, SIDE side );
int SendOk( CLIENTDATA * p, PIECE gotten, int isCheck );
int SendNext( CLIENTDATA * p, int isCaptured,
			  int CapturedX, int CapturedY, int isChecked );
int SendIllegal( CLIENTDATA *p );
int SendViolate( CLIENTDATA * p, VIOLATION kind_of_violation );
int SendEnd( CLIENTDATA * p, RESULT result, REASON_OF_END reason_of_end );
int SendGameList( CLIENTDATA * p );
int SendViewStart( CLIENTDATA * p );
int SendViewHandData( CLIENTDATA *p, const NOTATION_HANDDATA *data );
int SendCurrentHandsForViewer( CLIENTDATA *p );
int SendEachViewerHandData( GAMEDATA *game, NOTATION_HANDDATA *data );
int SendFile( CLIENTDATA *p, char *filename );
int SendNotation( CLIENTDATA *client, int NotationID );
int SendCheat( CLIENTDATA *client, SIDE side, PIECE piece, int isPromoted );


void HandDataToMsg( const NOTATION_HANDDATA *data, char *msg );

/* ** Func SendMessage ******************************** */
/*
 * Write queued data into socket.
 * return 1 --- error
 * return 0 --- success
 */
int SendMessage( CLIENTDATA *client )
{
/* -- Local Data --------------------------------- */
    SENDMSGDATA *pMsg;
    int writeret;
/* -- Operation ---------------------------------- */
    pMsg = client->SendMsgList;
    if( pMsg == NULL ) return 1;
    writeret = write( client->soc, pMsg->data, pMsg->datalen );
    if( writeret < 0 ) {
        Err( "write." );
        return 1;
    } else if( writeret == pMsg->datalen ) {
        client->SendMsgList = pMsg->Next;
        free( pMsg->data );
        free( pMsg );
        if( client->SendMsgList == NULL && client->ShouldBeRemoved ) {
            RemoveClient( client );
        }
    } else {
        char *p;
        p = malloc( pMsg->datalen - writeret );
        if( p == NULL ) {
            Err( "malloc." );
            return 1;
        }
        memcpy( p, pMsg->data + writeret, pMsg->datalen - writeret );
        pMsg->datalen -= writeret;
    }
    return 0;
} /* SendMessage */


/* ** Func SendStr ******************************* */
/*
 * Add given String to message queue of client.
 * return 1 --- error
 * return 0 --- success
 */
int SendStr( CLIENTDATA *p, const char *str )
{
/* -- Local Data --------------------------------- */
    int ret;
/* -- Operation ---------------------------------- */
    ret = SendData( p, str, strlen( str ) );
    if( !ret ) 
        AccessLog( "to client %d '%s'", p->soc, str );
    return ret;
} /* SendStr */

/* ** Func SendData ******************************* */
/*
 * Add given data (binary) to message queue of client.
 * return 1 --- error
 * return 0 --- success
 */
int SendData( CLIENTDATA *p, const char *data, int datalen )
{
/* -- Local Data --------------------------------- */
    SENDMSGDATA *msg;
/* -- Operation ---------------------------------- */
    /* add new message to message list */
    msg = malloc( sizeof( SENDMSGDATA ) );
    if( msg == NULL ) {
        Err( "malloc." );
        return 1;
    }
    msg->Next = NULL;
    msg->datalen = datalen;
    msg->data = malloc( msg->datalen );
    if( msg->data == NULL ) {
        Err( "malloc." );
        free( msg );
        return 1;
    }
    memcpy( msg->data, data, msg->datalen );
    if( p->SendMsgList == NULL ) {
        p->SendMsgList = msg;
    } else {
        SENDMSGDATA *cur = p->SendMsgList;
        while( cur->Next != NULL )
            cur = cur->Next;
        cur->Next = msg;
    }
    return 0;
} /* SendData */

/* ** Func SendHelp ****************************** */
/*
 * Send help (usage).
 * return 1 --- error
 * return 0 --- success
 */
int SendHelp( CLIENTDATA * p )
{
/* -- Local Data --------------------------------- */
    const char HelpHeader[] =
        "<tuitated version 4.0>\n"
        "which is supporting TUITATE PROTOCOL version 4.\n"
	"Currennt supported commands.\n"
	"\tHELP\n"
	"\tMODE\n"
	"\tQUIT\n";
    char help[512];
/* -- Operation ---------------------------------- */
    strcpy( help, HelpHeader );
    switch( p->mode ) {
      case MODE_NEUTRAL:
        strcat( help, "\tGAME\n" );
        strcat( help, "\tVIEW\n" );
        strcat( help, "\tNOTATION\n" );
        strcat( help, "\tMANAGENOTATION\n" );
        strcat( help, "\tGET\n" );
        break;
      case MODE_SELECTENEMY:
        strcat( help, "\tCHALLENGERLIST\n" );
        strcat( help, "\tCHALLENGE\n" );
        strcat( help, "\tACCEPT\n" );
        strcat( help, "\tREFUSE\n" );
        break;
      case MODE_GAME:
        strcat( help, "\tMOVE\n" );
        strcat( help, "\tDROP\n" );
        strcat( help, "\tGIVEUP\n" );
        break;
      case MODE_SELECTGAME:
        strcat( help, "\tGAMELIST\n" );
        strcat( help, "\tVIEWGAME\n" );
        break;
      case MODE_VIEW:
      case MODE_GET:
        break;
      case MODE_SELECTNOTATION:
        strcat( help, "\tNOTATIONLIST\n" );
        strcat( help, "\tVIEWNOTATION\n" );
        break;
      case MODE_MANAGENOTATION:
        strcat( help, "\tNOTATIONLIST\n" );
        strcat( help, "\tDELETENOTATION\n" );
        strcat( help, "\tSAVENOTATION\n" );
        strcat( help, "\tUNSAVENOTATION\n" );
        break;
    }
    strcat( help, ".\n" );
    return SendStr( p, help );
} /* SendHelp */

/* ** Func SendMode ****************************** */
/*
 * Send current mode of the client.
 * return 1 --- error
 * return 0 --- success
 */
int SendMode( CLIENTDATA * p )
{
/* -- Local Data --------------------------------- */
    char mode[256];
/* -- Operation ---------------------------------- */
    strcpy( mode, "MODE " );
    switch( p->mode ) {
      case MODE_NEUTRAL:
        strcat( mode, "NEUTRAL" );
        break;
      case MODE_SELECTENEMY:
        strcat( mode, "SELECTENEMY" );
        break;
      case MODE_GAME:
        strcat( mode, "GAME" );
        break;
      case MODE_SELECTGAME:
        strcat( mode, "SELECTGAME" );
        break;
      case MODE_VIEW:
        strcat( mode, "VIEW" );
        break;
      case MODE_SELECTNOTATION:
        strcat( mode, "SELECTNOTATION" );
        break;
      case MODE_MANAGENOTATION:
        strcat( mode, "MANAGENOTATION" );
        break;
      case MODE_GET:
        strcat( mode, "GET" );
        break;
    }
    strcat( mode, "\n" );
    return SendStr( p, mode );
} /* SendMode */

/* ** Func SendStart ********************* */
/*
 * Send start  (and is Sente or Gote).
 */
int SendStart( CLIENTDATA * p, SIDE side )
{
/* -- Local Data --------------------------------- */
    char buf[64];
/* -- Operation ---------------------------------- */
    strcpy( buf, KEY_START );
    strcat( buf, " " );
    switch( side ) {
      case BLACK:
        strcat( buf, STR_BLACK );
        break;
      case WHITE:
        strcat( buf, STR_WHITE );
        break;
      default:
        /* should not come here... */
        break;
    }
    strcat( buf, "\n" );
    return SendStr( p, buf );
} /* SendStart */

/* ** Func SendChallengerlist ********************* */
/*
 * Send the list of clients
 * which is in SELECTENEMY mode.
 * return 1 --- error
 * return 0 --- success
 */
int SendChallengerlist( CLIENTDATA *p )
{
/* -- Local Data --------------------------------- */
    char buf[256];
    char tmp[64];
    CLIENTDATA *data;
/* -- Operation ---------------------------------- */
    if( p->mode != MODE_SELECTENEMY ) return 1;
    strcpy( buf, KEY_CHALLENGERLIST );
    strcat( buf, " " );
    sprintf( tmp, "%d", ChallengerNum( p ) );
    strcat( buf, tmp );
    strcat( buf, "\n" );
    if( SendStr( p, buf ) ) return 1;
    data = ClientList;
    while( data != NULL ) {
        if( data->mode == MODE_SELECTENEMY && data != p ) {
            sprintf( buf, "%d", data->soc );
            if( data->name != NULL ) {
                strcat( buf, " " );
                strcat( buf, data->name );
                if( data->mailaddress != NULL ) {
                    strcat( buf, " " );
                    strcat( buf, data->mailaddress );
                }
            }
            strcat( buf, "\n" );
            if( SendStr( p, buf ) ) return 1;
        }
        data = data->Next;
    }
    return SendStr( p, ".\n" );
} /* SendChallengerlist */

/* ** Func SendChallenged *************************** */
/*
 * Send the selected client is challenged
 * by another client.
 * Return 1 on Error.
 * Return o on Success.
 */
int SendChallenged( CLIENTDATA *p, CLIENTDATA *challenger )
{
/* -- Local Data --------------------------------- */
    char buf[256];
    char tmp[64];
/* -- Operation ---------------------------------- */
    if( p->mode != MODE_SELECTENEMY ) return 1;
    strcpy( buf, KEY_CHALLENGED );
    strcat( buf, " " );
    sprintf( tmp, "%d", challenger->soc );
    strcat( buf, tmp );
    if( challenger->name != NULL ) {
        strcat( buf, " " );
        strcat( buf, challenger->name );
        if( challenger->mailaddress != NULL ) {
            strcat( buf, " " );
            strcat( buf, challenger->mailaddress );
        }
    }
    strcat( buf, "\n" );
    return SendStr( p, buf );
} /* SendChallenged */

/* ** Func SendNotExist ************************** */
/*
 * Send that the selected client does not Exist now.
 * Return 1 on Error.
 * Return o on Success.
 */
int SendNotExist( CLIENTDATA * p )
{
/* -- Local Data --------------------------------- */
    char buf[64];
/* -- Operation ---------------------------------- */
    if( (p->mode != MODE_SELECTENEMY) &&
        (p->mode != MODE_SELECTNOTATION) &&
	(p->mode != MODE_SELECTGAME)) return 1;
    strcpy( buf, KEY_NOTEXIST );
    strcat( buf, "\n" );
    return SendStr( p, buf );
} /* SendNotExist */

/* ** Func SendBusy ****************************** */
/*
 * Send that the selected client is Busy to contact.
 * Return 1 on Error.
 * Return o on Success.
 */
int SendBusy( CLIENTDATA * p )
{
/* -- Local Data --------------------------------- */
    char buf[64];
/* -- Operation ---------------------------------- */
    if( p->mode != MODE_SELECTENEMY ) return 1;
    strcpy( buf, KEY_BUSY );
    strcat( buf, "\n" );
    return SendStr( p, buf );
} /* SendBusy */

/* ** Func SendRefused *************************** */
/*
 * Send that the selected client is Busy to contact.
 * Return 1 on Error.
 * Return o on Success.
 */
int SendRefused( CLIENTDATA * p )
{
/* -- Local Data --------------------------------- */
    char buf[64];
/* -- Operation ---------------------------------- */
    if( p->mode != MODE_SELECTENEMY ) return 1;
    strcpy( buf, KEY_REFUSED );
    strcat( buf, "\n" );
    return SendStr( p, buf );
} /* SendRefused */

/* ** Func SendEnemyViolate ********************* */
/*
 * Send that enemy has violated.
 * Return 1 on Error.
 * Return o on Success.
 */
int SendEnemyViolate( CLIENTDATA * p )
{
/* -- Local Data --------------------------------- */
    char buf[64];
/* -- Operation ---------------------------------- */
    if( p->mode != MODE_GAME ) return 1;
    strcpy( buf, KEY_ENEMYVIOLATE );
    strcat( buf, "\n" );
    return SendStr( p, buf );
} /* SendEnemyViolate */

/* ** Func SendOk ********************* */
/*
 * Send Ok (and 'koma gotten' and 'is Checking').
 * Return 1 on Error.
 * Return 0 on Success.
 */
int SendOk( CLIENTDATA * p, PIECE gotten, int isCheck )
{
/* -- Local Data --------------------------------- */
    char buf[64];
/* -- Operation ---------------------------------- */
    if( p->mode != MODE_GAME ) return 1;
    strcpy( buf, KEY_OK );
    if( gotten != NONE ) {
        strcat( buf, " " );
        strcat( buf, PieceStr[gotten] );
    }
    if( isCheck ) {
        strcat( buf, " " );
        strcat( buf, STR_CHECK );
    }
    strcat( buf, "\n" );
    return SendStr( p, buf );
} /* SendOk */

/* ** Func SendNext ********************* */
/*
 * Send Next( and isCheced ).
 * if Next Player's piece is Captured,
 * send CapturedX and CapturedY.( 1 - 9 )
 * if Not Captured, CapturedX and CapturedX should be 0.
 */
int SendNext( CLIENTDATA *p, int isCaptured, int CapturedX, int CapturedY, int isChecked )
{
/* -- Local Data --------------------------------- */
    char buf[64];
/* -- Operation ---------------------------------- */
    if( p->mode != MODE_GAME ) return 1;
    strcpy( buf, KEY_NEXT );
    if( isCaptured && 
        CapturedX >= 0 && CapturedX <= 8 && CapturedY >= 0 && CapturedY <= 8) {
        char tmpbuf[16];
        CapturedX++;
        CapturedY++;
        /*
        if( p == p->game->White ) {
            CapturedX = 10 - CapturedX;
            CapturedY = 10 - CapturedY;
        }
        */
        strcat( buf, " " );
        sprintf( tmpbuf, "%d %d", CapturedX, CapturedY );
        strcat( buf, tmpbuf );
    }
    if( isChecked ) {
        strcat( buf, " " );
        strcat( buf, STR_CHECKED );
    }
    strcat( buf, "\n" );
    return SendStr( p, buf );
} /* SendNext */

/* ** Func SendIllegal ********************* */
/*
 * Send Illegal.
 */
int SendIllegal( CLIENTDATA * p )
{
/* -- Local Data --------------------------------- */
    char buf[64];
/* -- Operation ---------------------------------- */
    if( p->mode != MODE_GAME ) return 1;
    strcpy( buf, KEY_ILLEGAL );
    strcat( buf, "\n" );
    return SendStr( p, buf );
} /* SendIllegal */

/* ** Func SendViolate ********************* */
/*
 * Send Violation and what kind of it.
 */
int SendViolate( CLIENTDATA * p, VIOLATION kind_of_violation )
{
/* -- Local Data --------------------------------- */
    char buf[64];
/* -- Operation ---------------------------------- */
    if( p->mode != MODE_GAME ) return 1;
    strcpy( buf, KEY_VIOLATE );
    strcat( buf, " " );
    strcat( buf, ViolationStr[kind_of_violation] );
    strcat( buf, "\n" );
    return SendStr( p, buf );
} /* SendViolate */

/* ** Func SendEndID ********************* */
/*
 * Send End ID line( Win or lose, and the Reason ).
 * return 1 --- ERROR
 * return 0 --- SUCCESS
 */
int SendEndID( CLIENTDATA * p, RESULT result, REASON_OF_END reason_of_end )
{
/* -- Local Data --------------------------------- */
    char buf[256];
    char tmp[64];
/* -- Operation ---------------------------------- */
    if( p->mode != MODE_GAME ) return 1;
    strcpy( buf, KEY_END );
    strcat( buf, " " );
    switch( result ) {
      case WIN:
        strcat( buf, STR_WIN );
        break;
      case LOSE:
        strcat( buf, STR_LOSE );
        break;
      case DRAW:
        strcat( buf, STR_DRAW );
        break;
    }
    strcat( buf, " " );
    strcat( buf, ReasonOfEndStr[reason_of_end] );
    strcat( buf, " " );
    sprintf( tmp, "%d", HandNum( p->game ) );
    strcat( buf, tmp );
	
    strcat( buf, "\n" );
    return SendStr( p, buf );
} /* SendEndID */


/* ** Func SendEndData ********************* */
/*
 * Send End Data lines.
 * return 1 --- ERROR
 * return 0 --- SUCCESS
 */
int SendEndData( CLIENTDATA * p, GAMEDATA *game )
{
/* -- Local Data --------------------------------- */
    NOTATION_HANDDATA *pData;
    NOTATION_HANDDATA data;
    char buf[256];
    char tmp[64];
/* -- Operation ---------------------------------- */
    pData = game->HandList;
    while( pData != NULL ) {
        memcpy( &data, pData, sizeof( NOTATION_HANDDATA ) );
        switch( data.key ) {
          case NOTATION_MOVE:
            strcpy( buf, KEY_MOVE );
            strcat( buf, " " );
            tmp[0] = '1' + data.hand.move.sx; tmp[1] = '\0';
            strcat( buf, tmp );
            strcat( buf, " " );
            tmp[0] = '1' + data.hand.move.sy; tmp[1] = '\0';
            strcat( buf, tmp );
            strcat( buf, " " );
            tmp[0] = '1' + data.hand.move.dx; tmp[1] = '\0';
            strcat( buf, tmp );
            strcat( buf, " " );
            tmp[0] = '1' + data.hand.move.dy; tmp[1] = '\0';
            strcat( buf, tmp );
            strcat( buf, " " );
            if( data.hand.move.willPromote ) {
                strcat( buf, STR_PROMOTE );
            } else {
                strcat( buf, STR_NOTPROMOTE );
            }
            strcat( buf, " " );
            if( data.isChecked ) {
                strcat( buf, STR_CHECKED );
            } else {
                strcat( buf, STR_NOTCHECKED );
            }
            if( data.isCaptured ) {
                strcat( buf, " " );
                strcat( buf, PieceStr[data.capture.piece] );
                strcat( buf, " " );
                if( data.capture.hasPromoted ) {
                    strcat( buf, STR_PROMOTE );
                } else {
                    strcat( buf, STR_NOTPROMOTE );
                }
            }
            if( data.isViolate ) {
                strcat( buf, " " );
                strcat( buf, ViolationStr[data.violation] );
            }
            break;
          case NOTATION_DROP:
            strcpy( buf, KEY_DROP );
            strcat( buf, " " );
            tmp[0] = '1' + data.hand.drop.dx; tmp[1] = '\0';
            strcat( buf, tmp );
            strcat( buf, " " );
            tmp[0] = '1' + data.hand.drop.dy; tmp[1] = '\0';
            strcat( buf, tmp );
            if( data.hand.drop.piece != NONE ) {
                strcat( buf, " " );
                strcat( buf, PieceStr[data.hand.drop.piece] );
            }
            strcat( buf, " " );
            if( data.isChecked ) {
                strcat( buf, STR_CHECKED );
            } else {
                strcat( buf, STR_NOTCHECKED );
            }
            if( data.isViolate ) {
                strcat( buf, " " );
                strcat( buf, ViolationStr[data.violation] );
            }
            break;
          case NOTATION_END:
            strcpy( buf, KEY_END );
            strcat( buf, " " );
            switch( data.winner ) {
              case BLACK:
                strcat( buf, STR_BLACK );
                break;
              case WHITE:
                strcat( buf, STR_WHITE );
                break;
              case NOSIDE:
                strcat( buf, STR_NOSIDE );
                break;
            }
            strcat( buf, " " );
            strcat( buf, ReasonOfEndStr[data.reason_of_end] );
            break;
        }
        strcat( buf, "\n" );
        if( SendStr( p, buf ) ) return 1;
        pData = pData->Next;
    }
    strcpy( buf, ".\n" );
    return SendStr( p, buf );
} /* SendEndData */

/* ** Func SendGameList ********************* */
/*
 * Send GameList.
 */
int SendGameList( CLIENTDATA * p )
{
/* -- Local Data --------------------------------- */
    GAMEDATA *game;
    char tmp[64];
    char buf[256];
/* -- Operation ---------------------------------- */
    strcpy( buf, KEY_GAMELIST );
    strcat( buf, " " );
    sprintf( tmp, "%d", GameNum() );
    strcat( buf, tmp );
    strcat( buf, "\n" );
    if( SendStr( p, buf ) ) return 1;
    for( game = GameList; game != NULL; game = game->Next ) {
        sprintf( tmp, "%d", game->id );
        strcpy( buf, tmp );
        strcat( buf, " " );
        if( game->Black->name != NULL ) {
            strcat( buf, game->Black->name );
        } else {
            strcat( buf, STR_UNKNOWN );
        }
        strcat( buf, " " );
        if( game->Black->mailaddress != NULL ) {
            strcat( buf, game->Black->mailaddress );
        } else {
            strcat( buf, STR_UNKNOWN );
        }
        strcat( buf, " " );
        if( game->White->name != NULL ) {
            strcat( buf, game->White->name );
        } else {
            strcat( buf, STR_UNKNOWN );
        }
        strcat( buf, " " );
        if( game->White->mailaddress != NULL ) {
            strcat( buf, game->White->mailaddress );
        } else {
            strcat( buf, STR_UNKNOWN );
        }
        strcat( buf, "\n" );
        if( SendStr( p, buf ) ) return 1;
    }
    strcpy( buf, ".\n" );
    return SendStr( p, buf );
} /* SendGameList */

/* ** Func SendViewStart ********************* */
/*
 * Send ViewStart.
 */
int SendViewStart( CLIENTDATA * p )
{
/* -- Local Data --------------------------------- */
    char buf[64];
/* -- Operation ---------------------------------- */
    strcpy( buf, KEY_VIEWSTART );
    strcat( buf, "\n" );
    return SendStr( p, buf );
} /* SendViewStart */

/* ** Func SendViewHandData ********************** */
/*
 * Send Viewer Notation Hand Data.
 */
int SendViewHandData( CLIENTDATA *p, const NOTATION_HANDDATA *data )
{
/* -- Local Data --------------------------------- */
    char buf[256];
/* -- Operation ---------------------------------- */
    HandDataToMsg( data, buf );
    return  SendStr( p, buf );
} /* SendViewHandData */

/* ** Func HandDataToMsg *************************** */
/*
 * Convert HandData into Messsage String ( with return code ).
 * Given message data field must be greater than 256 byte.
 */
void HandDataToMsg( const NOTATION_HANDDATA *data, char *msg )
{
/* -- Local Data --------------------------------- */
    char tmp[64];
/* -- Operation ---------------------------------- */
    switch( data->key ) {
      case NOTATION_MOVE:
        strcpy( msg, KEY_MOVE );
        strcat( msg, " " );
        sprintf( tmp, "%d", data->hand.move.sx + 1 );
        strcat( msg, tmp );
        strcat( msg, " " );
        sprintf( tmp, "%d", data->hand.move.sy + 1 );
        strcat( msg, tmp );
        strcat( msg, " " );
        sprintf( tmp, "%d", data->hand.move.dx + 1 );
        strcat( msg, tmp );
        strcat( msg, " " );
        sprintf( tmp, "%d", data->hand.move.dy + 1 );
        strcat( msg, tmp );
        strcat( msg, " " );
        if( data->hand.move.willPromote ) {
            strcat( msg, STR_PROMOTE );
        } else {
            strcat( msg, STR_NOTPROMOTE );
        }
        strcat( msg, " " );
        if( data->isChecked ) {
            strcat( msg, STR_CHECKED );
        } else {
            strcat( msg, STR_NOTCHECKED );
        }
        if( data->isCaptured ) {
            strcat( msg, " " );
            strcat( msg, PieceStr[data->capture.piece] );
            strcat( msg, " " );
            strcat( msg,
                    data->capture.hasPromoted ? STR_PROMOTE : STR_NOTPROMOTE );
        }
        if( data->isViolate ) {
            strcat( msg, " " );
            strcat( msg, ViolationStr[data->violation] );
        }
        break;
      case NOTATION_DROP:
        strcpy( msg, KEY_DROP );
        strcat( msg, " " );
        sprintf( tmp, "%d", data->hand.drop.dx + 1 );
        strcat( msg, tmp );
        strcat( msg, " " );
        sprintf( tmp, "%d", data->hand.drop.dy + 1 );
        strcat( msg, tmp );
        strcat( msg, " " );
        strcat( msg, PieceStr[data->hand.drop.piece] );
        strcat( msg, " " );
        if( data->isChecked ) {
            strcat( msg, STR_CHECKED );
        } else {
            strcat( msg, STR_NOTCHECKED );
        }
        if( data->isViolate ) {
            strcat( msg, " " );
            strcat( msg, ViolationStr[data->violation] );
        }
        break;
      case NOTATION_END:
        strcpy( msg, KEY_END );
        strcat( msg, " " );
        switch( data->winner ) {
          case BLACK:
            strcat( msg, STR_BLACK );
            break;
          case WHITE:
            strcat( msg, STR_WHITE );
            break;
          case NOSIDE:
            strcat( msg, STR_NOSIDE );
            break;
        }
        strcat( msg, " " );
        strcat( msg, ReasonOfEndStr[data->reason_of_end] );
        break;
    }
    strcat( msg, "\n" );
} /* HandDataToMsg */

/* ** Func SendCurrentHandsForViewer ************** */
/*
 * Send Current Saved Hands for given Viewer Client.
 */
int SendCurrentHandsForViewer( CLIENTDATA *p )
{
/* -- Local Data --------------------------------- */
    NOTATION_HANDDATA *hand;
/* -- Operation ---------------------------------- */
    if( p->mode != MODE_VIEW ) return 1;
    hand = p->game->HandList;
    while( hand != NULL ) {
        if( SendViewHandData( p, hand ) ) return 1;
        hand = hand->Next;
    }
    return 0;
} /* SendCurrentHandsForViewer */

/* ** Func SendEachViewerHandData **************** */
/*
 * Send the given hand to all of the viewer of
 * given Game.
 */
int SendEachViewerHandData( GAMEDATA *game, NOTATION_HANDDATA *data )
{
/* -- Local Data --------------------------------- */
    VIEWERDATA *cur;
    VIEWERDATA *prev;
/* -- Operation ---------------------------------- */
    prev = NULL;
    cur = game->ViewerList;
    while( cur != NULL ) {
        if( SendViewHandData( cur->client, data ) ) {
            /* if cannot send message to viewer, remove the viewer */
            if( prev == NULL ) {
                game->ViewerList = cur->Next;
                DeleteClient( cur->client );
                free( cur );
                cur = game->ViewerList;
            } else {
                prev->Next = cur->Next;
                DeleteClient( cur->client );
                free( cur );
                cur = prev->Next;
            }
        } else {
            prev = cur;
            cur = cur->Next;
        }
    }
    return 0;
} /* SendEachViewerHandData */

/* ** Func SendFile **************** */
/*
 * Send the given file.
 * Given filename contains '/' at head of it.
 * return 0 if success.
 * return 1 if given file is not found.
 */
int SendFile( CLIENTDATA *p, char *filename )
{
/* -- Local Data --------------------------------- */
    const char HeadMessage[] = "HTTP/1.0 200 OK\n";
    const char TextMessage[] = "Content-Type: text/html\n";
    const char ApplMessage[] = "Content-Type: application/octet-stream\n";
    const char GifMessage[] = "Content-Type: image/gif\n";
    const char *TypeMessage[] = { TextMessage, ApplMessage, GifMessage };
    const char SizeMessage[] = "Content-Length: ";

    char buf[1024];
    char c;
    FILE *fp;
    int i;
    int filetype = 0;
/* -- Operation ---------------------------------- */
    strcpy( buf, ExportDir );
    if( !strcmp( filename, "/" ) ) {
        strcat( buf, "/index.html" );
    } else {
        strcat( buf, filename );
    }
    for( i = strlen( filename ) - 1; i > 0; i-- ) {
        if( filename[i] == '.' ) {
            break;
        }
    }
    if( i != 0 ) {
        if( !strcmp( filename + i + 1, "class" ) ) {
            filetype = 1;
        } else if( !strcmp( filename + i + 1, "gif" ) ) {
            filetype = 2;
        }
    }
    fp = fopen( buf, "r" );
    if( fp == NULL ) {
        return 1;
    } else {
        int size = 0;
        int FreadRet;
        AccessLog( "to client %d  (sendfile %s)", p->soc, buf );
        ServiceLog( "SendFile %s to client %d", buf, p->soc );
        SendStr( p, HeadMessage );
        SendStr( p, TypeMessage[filetype] );
        while( fread( &c, 1, 1, fp ) ) {
            size++;
        }
        sprintf( buf, "%s%d\n\n", SizeMessage, size );
        SendStr( p, buf );
        fseek( fp, 0, SEEK_SET );
        while( (FreadRet = fread( buf, 1, sizeof( buf ), fp )) > 0 ) {
            SendData( p, buf, FreadRet );
        }
        fclose( fp );
    }
    return 0;
} /* SendFile */

/* ** Func SendNotationList ******************** */
/*
 * Send list of notations to client.
 * If given 'toManager' flag is TRUE,
 * property of notations will be sent too.
 */
int SendNotationList( CLIENTDATA *client, int toManager )
{
/* -- Local Data --------------------------------- */
    NOTATIONDATA *p;
    char buf[256];
/* -- Operation ---------------------------------- */
    /* UpdateNotationList(); */
    sprintf( buf, "%s %d\n", KEY_NOTATIONLIST, NotationNum() );
    if( SendStr( client, buf ) ) return 1;
    p = NotationList;
    while( p != NULL ) {
        sprintf( buf, "%d %ld %ld", p->ID, p->Begin, p->End );
        if( p->saved && toManager ) {
            strcat( buf, " " );
            strcat( buf, STR_SAVE );
        }
        strcat( buf, "\n" );
        if( SendStr( client, buf ) ) return 1;
        sprintf( buf, "%s %s %s %s\n", p->BlackName, p->BlackAddress,
                p->WhiteName, p->WhiteAddress );
        if( SendStr( client, buf ) ) return 1;
        p = p->Next;
    }
    return SendStr( client, ".\n" );
} /* SendNotationList */

/* ** Func SendNotation *************************** */
/*
 * Send Notation data.
 * If an error occurs, return 1.
 * If no error occurs, return 0.
 */
int SendNotation( CLIENTDATA *client, int NotationID )
{
/* -- Local Data --------------------------------- */
    NOTATIONDATA *cur;
    char buf[512];
    int NotationLineNum = 0;
    int isNotationSentComplete = FALSE;
    int isNotationSentStarted = FALSE;
/* -- Operation ---------------------------------- */
Err( "in SendNotaion. (NotationID = %d)\n", NotationID );
    cur = NotationList;
    while( cur != NULL ) {
Err( "in SendNotaion Loop. ( loop... cur id = %d)\n", cur->ID );
        if( cur->ID == NotationID ) break;
        cur = cur->Next;
    }
    if( cur != NULL ) { /* notation found! */
        FILE *fp;
        fp = fopen( NotationFile, "r" );
        if( fp == NULL ) {
            Err( "Notation File %s cannot open.", NotationFile );
            return 1;
        }
        fseek( fp, cur->NotationPos, 0 );
        for( ;; ) {
            int i;
            /* count notation line num */
            while( 1 ) {
                if( fgets( buf, 256+1, fp ) == NULL ) {
                    fclose( fp );
                    return 1; /* read error */
                } else if( !strncmp( buf, KEY_END, strlen(KEY_END) ) ) {
                    NotationLineNum++;
                    break;
                }
                NotationLineNum++;
            }
            fseek( fp, cur->NotationPos, 0 );
            sprintf( buf, "NOTATION %d\n", NotationLineNum );
            if( SendStr( client, buf ) ) {
                break;       /* write error */
            }
            isNotationSentStarted = TRUE;
            /* send notation */
            for( i = 0; i < NotationLineNum; i++ ) {
                if( fgets( buf, 256+1, fp ) == NULL ) {
                    break;
                }
                if( SendStr( client, buf ) ) {
                    break;       /* write error */
                }
            }
            ServiceLog( "Client %d view notation.(%s vs %s)",
                        client->soc, cur->BlackName, cur->WhiteName );
            isNotationSentComplete = TRUE;
            break;
        }
        fclose( fp );
        if( !isNotationSentComplete ) {
            if( isNotationSentStarted ) {
                SendStr( client, ".\n" );
            }
            return 1;
        }
        if( SendStr( client, ".\n" ) ) return 1;
    } else { /* Notation NOT found ! */
        return 1;
    }
    return 0;
} /* SendNotation */

/* ** Func SendCheat *************************** */
/*
 * Send Notation data.
 * If an error occurs, return 1.
 * If no error occurs, return 0.
 */
int SendCheat( CLIENTDATA *client, SIDE side, PIECE piece, int isPromoted )
{
/* -- Local Data --------------------------------- */
    char buf[512];
/* -- Operation ---------------------------------- */
    sprintf( buf, "%s ", KEY_CHEAT );
    switch( side ) {
      case BLACK:
        strcat( buf, STR_BLACK );
        break;
      case WHITE:
        strcat( buf, STR_WHITE );
        break;
      case NOSIDE:
        strcat( buf, STR_NONE );
        strcat( buf, "\n" );
        if( SendStr( client, buf ) ) return 1;
        return 0;
    }
    strcat( buf, " " );
    strcat( buf, PieceStr[piece] );
    if( isPromoted ) {
        strcat( buf, " " );
        strcat( buf, STR_PROMOTE );
    } else {
        strcat( buf, " " );
        strcat( buf, STR_NOTPROMOTE );
    }
    strcat( buf, "\n" );
    return SendStr( client, buf );
} /* SendCheat */


