/* ** struct.h *********************************** */
/* Struct definitions for tuitated. */
/* *********************************************** */
#ifndef _STRUCT_H_
#define _STRUCT_H_

/* -- Include File ------------------------------- */
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include "protocol.h"
#include "tuitated.h"

/* -- Macro Definition --------------------------- */

/* -- Type Decleration --------------------------- */
typedef struct _CLIENTDATA *PCLIENTDATA;
typedef struct _VIEWERDATA *PVIEWERDATA;
typedef struct _GAMEDATA *PGAMEDATA;
typedef struct _NOTATION_HANDDATA *PNOTATION_HANDDATA;
typedef struct _NOTATIONDATA *PNOTATIONDATA;
typedef struct _SENDMSGDATA *PSENDMSGDATA;

#define CLIENT_BUFSIZE 512

typedef struct _CLIENTDATA {
    PCLIENTDATA Next;
    int soc;
    MODETYPE mode;
    PGAMEDATA game;
    char *name;
    char *mailaddress;
    PCLIENTDATA Selecting;
    PCLIENTDATA Selected;
    char ReceiveMsgBuf[CLIENT_BUFSIZE];
    PSENDMSGDATA SendMsgList;
    int ShouldBeRemoved;    /* if this flag is TRUE and SendMsgList is empty,
                               connection to this client should be closed */
} CLIENTDATA;

typedef struct _VIEWERDATA {
    PVIEWERDATA Next;
    PCLIENTDATA client;
} VIEWERDATA;

typedef struct _SENDMSGDATA {
    PSENDMSGDATA Next;
    char *data;
    int datalen;
} SENDMSGDATA;

typedef struct _BOX {
    CLIENTDATA *Belong;
    PIECE PieceType;
    int isPromoted;
} BOX;

typedef struct _PIECES_IN_HAND {
    int Pawn;
    int Lance;
    int Knight;
    int Silver;
    int Gold;
    int Bishop;
    int Rook;
} PIECES_IN_HAND;

typedef struct _GAMEDATA {
    int			id;
    PGAMEDATA		Next;
    time_t              Begin;  /* time when this game started */
    CLIENTDATA		*Black;
    CLIENTDATA		*White;
    CLIENTDATA		*CurrentPlayer;
    CLIENTDATA		*NonCurrentPlayer;
    int			ActionNumber;
    BOX			Board[9][9];
    PIECES_IN_HAND	BlackPiecesInHand;
    PIECES_IN_HAND	WhitePiecesInHand;
    int			BlackLifeNum;
    int			WhiteLifeNum;
    PNOTATION_HANDDATA	HandList;
    PVIEWERDATA		ViewerList;
} GAMEDATA;

typedef enum {
    OK,
    ILLEGAL,
    VIOLATE,
    END,
} JUDGERESULT;

typedef struct _ENDSTATE {
    CLIENTDATA		*winner;
    REASON_OF_END	reason;
} ENDSTATE;

typedef struct _CAPTURESTATE {
    PIECE piece;
    int x;
    int y;
    int hasPromoted;
} CAPTURESTATE;

typedef struct _JUDGE {
    JUDGERESULT result;
    int		isCheck;	/* result == OK */
    int		isCaptured;	/* result == OK */
    CAPTURESTATE capture;	/* isCaptured */
    VIOLATION	violation;	/* result == VIOLATE */
    ENDSTATE	endstate;	/* result == END */
} JUDGE;

typedef enum _NOTAION_KEY {
    NOTATION_MOVE,
    NOTATION_DROP,
    NOTATION_END,
} NOTATION_KEY;
							   
typedef struct _NOTATION_HANDDATA {
    PNOTATION_HANDDATA Next;
    NOTATION_KEY key;
    HANDDATA hand;
    int isCaptured;
    CAPTURESTATE capture;
    int isChecked;
    int isViolate;
    VIOLATION violation;
    SIDE winner;
    REASON_OF_END reason_of_end;
} NOTATION_HANDDATA;

typedef struct _NOTATIONDATA {
    PNOTATIONDATA Next;
    int ID;
    long NotationPos;  /* bytes count from head of notation file. */
    time_t Begin;
    time_t End;
    char *BlackName;
    char *BlackAddress;
    char *WhiteName;
    char *WhiteAddress;
    int saved;
} NOTATIONDATA;

#endif _STRUCT_H_

