/* ** tnls.c ************************************* */
/* Tuitate Notaiont LiSt. */
/* Management command.    */
/* *********************************************** */

/* -- Include File ------------------------------- */
#include <config.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

/* -- Global Data -------------------------------- */
unsigned short port;
char servername[256];

int ShowMail;
int ShowDate;
int ShowTime;

#define TRUE 1
#define FALSE 0

#define BUFSIZE 512

/* -- Prototype Declerations  -------------------- */

/* ** Func ConnectServer ************************* */
/*
 * Setup connection to given server and port.
 * Return socket of the connection.
 * On error, return -1.
 */
int ConnectServer( char *servname, unsigned short portnum )
{
/* -- Local Data --------------------------------- */
    struct hostent *servhost;
    struct sockaddr_in server;
    int s;
/* -- Operation ---------------------------------- */
    servhost = gethostbyname( servname );
    if( servhost == NULL ) {
        fprintf( stderr, "gethostbyname.\n" );
        return -1;
    }
    bzero( (char *)&server, sizeof(server) );
    server.sin_family = AF_INET;
    server.sin_port = htons( portnum );
    bcopy( servhost->h_addr, (char *)&server.sin_addr, servhost->h_length );
    if( (s = socket(AF_INET, SOCK_STREAM, 0)) < 0 ) {
        fprintf( stderr, "socket.\n" );
        return -1;
    }
    if( connect( s, (struct sockaddr *)&server, sizeof(server) ) == -1 ) {
        fprintf(stderr, "cannot connect.\n");
        return -1;
    }
    return s;
} /* ConnectServer */

/* ** Func  ********************************** */
/* Show SetDefault. */
void SetDefault( void )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
    port = PORT;
    strcpy( servername, "localhost" );
    ShowMail = FALSE;
    ShowDate = FALSE;
    ShowTime = FALSE;
} /* SetDefault */

/* ** Func Usage ********************************** */
/* Show Usage . */
void Usage( void )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
    printf( "Usage : tnls [switches]\n"
            "switches:\n"
            "   -d          display game dates.\n"
            "   -h          display help.\n"
            "   -l          display with long format.\n"
            "   -m          display mail addresses of players.\n"
            "   -p port     set port number to connect.\n"
            "   -s server   specify server to connect.\n"
            "   -t          display how long the games were played.\n"
        );
} /* Usage */

/* ** Func mystrtoi ****************************** */
/*
 * Convert string to integer.
 * Illegal input is not concerned.
 */
int mystrtoi( char *str )
{
/* -- Local Data --------------------------------- */
    int ret;
    int len;
    int base;
    int i;
    char *p;
/* -- Operation ---------------------------------- */
    ret = 0;
    for( p = str; *p != '\0'; p++ ) {
        len = strlen( p );
        for( i = 1, base = 1; i < len; i++ ) base *= 10;
        ret += base * ( *p - '0' );
    }
    return ret;
} /* mystrtoi */

/* ** Func ParseArg ********************************** */
/* Parse command line arguments. */
void ParseArg( int argc, char **argv )
{
/* -- Local Data --------------------------------- */
    int i;
    int ExpectPort = FALSE;
    int ExpectServer = FALSE;
/* -- Operation ---------------------------------- */
    for( i = 1; i < argc; i++ ) {
        if( ExpectServer ) {
            strcpy( servername, argv[i] );
            ExpectServer = FALSE;
        } else if( ExpectPort ) {
            port = mystrtoi( argv[i] );
            ExpectPort = FALSE;
        } else if( argv[i][0] == '-' ) {
            switch( argv[i][1] ) {
              case 'd':
                ShowDate = TRUE;
                break;
              case 'h':
                Usage();
                exit( 0 );
              case 'l':
                ShowMail = TRUE;
                ShowDate = TRUE;
                ShowTime = TRUE;
                break;
              case 'm':
                ShowMail = TRUE;
                break;
              case 'p':
                ExpectPort = TRUE;
                break;
              case 's':
                ExpectServer = TRUE;
                break;
              case 't':
                ShowTime = TRUE;
                break;
            }
        }
    }
} /* ParseArg */

/* ** Func ParamNum ****************************** */
/*
 * Return number of params in given string line.
 */
int ParamNum( char *line )
{
/* -- Local Data --------------------------------- */
    int isSpace; /* current examining char is space */
    char ret;
    char *p;
/* -- Operation ---------------------------------- */
    isSpace = TRUE;
    ret = FALSE;
    p = line;
    while( 1 ) {
        if( *p == '\0' ) {
            break;
        } else if( isspace( *p ) ) {
            if( !isSpace ) {
                isSpace = TRUE;
            }
        } else {
            if( isSpace ) {
                ret++;
                isSpace = FALSE;
            }
        }
        p++;
    }
    return ret;
} /* ParamNum */

/* ** Func GetParam ****************************** */
/*
 * Get One PARAM of given line( copy to given buf ),
 * which is nth ( 0, 1, ... ) in the line.
 * Return 0 .. OK.
 * Return 1 .. Error.
 */
int GetParam( char *line, int nth, char *buf )
{
/* -- Local Data --------------------------------- */
    int nthParam = 0;
    int isSpace = TRUE;
    int i = 0;
    char *p;
/* -- Operation ---------------------------------- */
    p = line;
    if( nth < 0 ) return 1;
    /* skip until nth PARAM appear */
    while( 1 ) {
        if( *p == '\0' ) {
            return 1;
        } else if( isspace( *p ) ) {
            if( !isSpace ) {
                isSpace = TRUE;
                nthParam++;
            }
        } else {
            if( isSpace ) {
                if( nth == nthParam ) {
                    break;
                } else {
                    isSpace = FALSE;
                }
            }
        }
        p++;
    }
    /* copy nth PARAM to given buffer */
    for( i = 0; ; i++ ) {
        buf[i] = toupper( p[i] );
        if( buf[i] == '\0') break;
        if( isspace( buf[i] ) ) {
            buf[i] = '\0';
            break;
        }
    }
    return 0;
} /* GetParam */
	
/* ** Func GetParamNotConverted ******************* */
/*
 * Get One PARAM of given line( copy to given buf ),
 * which is nth ( 0, 1, ... ) in the line.
 * This func is different from the func GetParam
 * on the point that this func do not convert chars
 * into upper case.
 * Return 0 .. OK.
 * Return 1 .. Error.
 */
int GetParamNotConverted( char *line, int nth, char *buf )
{
/* -- Local Data --------------------------------- */
    int nthParam = 0;
    int isSpace = TRUE;
    int i = 0;
    char *p;
/* -- Operation ---------------------------------- */
    p = line;
    if( nth < 0 ) return 1;
    /* skip until nth PARAM appear */
    while( 1 ) {
        if( *p == '\0' ) {
            return 1;
        } else if( isspace( *p ) ) {
            if( !isSpace ) {
                isSpace = TRUE;
                nthParam++;
            }
        } else {
            if( isSpace ) {
                if( nth == nthParam ) {
                    break;
                } else {
                    isSpace = FALSE;
                }
            }
        }
        p++;
    }
    /* copy nth PARAM to given buffer */
    for( i = 0; ; i++ ) {
        buf[i] = p[i];
        if( buf[i] == '\0') break;
        if( isspace( buf[i] ) ) {
            buf[i] = '\0';
            break;
        }
    }
    return 0;
} /* GetParamNotConverted */

/* ** Func main ********************************** */
/* main func */
int main( int argc, char **argv )
{
/* -- Local Data --------------------------------- */
    int soc;
    const char message[] = "MANAGENOTATION\n"
                           "NOTATIONLIST\n";
    char buf[BUFSIZE + 1];
    char *p;
    int ParsingDataLines = FALSE;
    int Getting1stLine = FALSE;
    int ID = 0;
    char BlackName[65];
    char BlackAddr[65];
    char WhiteName[65];
    char WhiteAddr[65];
    time_t StartTime = 0;
    time_t EndTime = 0;
    char StartTimeStr[64];
    char EndTimeStr[64];
    int saved = FALSE;
/* -- Operation ---------------------------------- */
    SetDefault();
    ParseArg( argc, argv );
    soc = ConnectServer( servername, port );
    if( soc < 0 ) {
        fprintf( stderr, "cannot connect to server.\n" );
        return 1;
    }
    write( soc, message, strlen( message ) );
    buf[0] = '\0';
    while( 1 ) {
        fd_set FdSet;
        int selectret;
        int readret;
        int datalen;
        int ShouldBreak = FALSE;
        FD_ZERO( &FdSet );
        FD_SET( soc, &FdSet );
        selectret = select( soc+1, &FdSet, NULL, NULL, NULL );
        if( selectret < 0 ) {
            fprintf( stderr, "select.\n" );
            break;
        }
        datalen = strlen( buf );
        readret = read( soc, buf + datalen, BUFSIZE - datalen );
        if( readret < 0 ) {
            fprintf( stderr, "read.\n" );
            break;
        }
        buf[readret + datalen] = '\0';
        while( 1 ) {
            p = strchr( buf, '\n' );
            if( p == NULL ) {
                if( strlen( buf ) > 256 ) {
                    fprintf( stderr, "Invalid message from server.\n" );
                    ShouldBreak = TRUE;
                    break;
                }
                break;
            } else {
                char line[512];
                char tmp[BUFSIZE + 1];
                *p = '\0';
                strcpy( line, buf );
                p++;
                strcpy( tmp, p );
                strcpy( buf, tmp );
                if( strlen( line ) > 256 ) {
                    fprintf( stderr, "Invalid message from server.\n" );
                    ShouldBreak = TRUE;
                    break;
                } else {
                    if( !strcmp( line, "." ) ) {
                        ShouldBreak = TRUE;
                        break;
                    }
                    if( !ParsingDataLines ) {
                        if( strncmp( line, "NOTATIONLIST", 12 ) ) {
                            fprintf( stderr, "Invalid message from server.\n");
                            ShouldBreak = TRUE;
                            break;
                        } else {
                            ParsingDataLines = TRUE;
                            Getting1stLine = TRUE;
                            GetParam( line, 1, tmp );
                            printf( "total %s\n", tmp );
                        }
                    } else if( Getting1stLine ) {
                        int num = ParamNum( line );
                        if( num != 3 && num != 4 ) {
                            fprintf( stderr, "Invalid message from server.\n");
                            ShouldBreak = TRUE;
                            break;
                        }
                        GetParam( line, 0, tmp );
                        ID = mystrtoi( tmp );
                        GetParam( line, 1, tmp );
                        StartTime = mystrtoi( tmp );
                        strftime( StartTimeStr, 63, "%Y %b %d %H:%M",
                                 localtime(&StartTime) );
                        GetParam( line, 2, tmp );
                        EndTime = mystrtoi( tmp );
                        strftime(EndTimeStr, 63, "%Y %b %d %H:%M",
                                 localtime(&EndTime) );
                        Getting1stLine = FALSE;
                        saved = FALSE;
                        if( num == 4 ) {
                            GetParam( line, 3, tmp );
                            if( !strcmp( tmp, "SAVE") ) {
                                saved = TRUE;
                            }
                        }
                    } else { /* Getting 2nd Line */
                        if( ParamNum( line ) != 4 ) {
                            fprintf( stderr, "Invalid message from server.\n");
                            ShouldBreak = TRUE;
                            break;
                        }
                        GetParamNotConverted( line, 0, tmp );
                        strcpy( BlackName, tmp );
                        GetParamNotConverted( line, 1, tmp );
                        strcpy( BlackAddr, tmp );
                        GetParamNotConverted( line, 2, tmp );
                        strcpy( WhiteName, tmp );
                        GetParamNotConverted( line, 3, tmp );
                        strcpy( WhiteAddr, tmp );
                        /* show */
                        printf( "%4d", ID );
                        if( saved )
                            printf( " SAVE " );
                        else
                            printf( "      " );
                        printf( "   %s", BlackName );
                        if( ShowMail )
                            printf( " (%s)", BlackAddr );
                        printf( " %s", WhiteName );
                        if( ShowMail )
                            printf( " (%s)", WhiteAddr );
                        if( ShowDate )
                            printf( " %s - %s", StartTimeStr, EndTimeStr );
                        if( ShowTime )
                            printf( " (%ld minutes)", (EndTime-StartTime)/60 );
                        printf( "\n" );
                        Getting1stLine = TRUE;
                    }
                    /*                    printf( "--- %s\n", line ); */
                }
            }
        }
        if( ShouldBreak ) break;
    }
    write( soc, "QUIT\n", 5 );
    close( soc );
    return 0;
} /* main */
