/* ** tnrm.c ************************************* */
/* Tuitate Notaiont ReMove. */
/* Management command.    */
/* *********************************************** */

/* -- Include File ------------------------------- */
#include <config.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

/* -- Global Data -------------------------------- */
unsigned short port;
char servername[256];

#define TRUE 1
#define FALSE 0

#define BUFSIZE 512

/* -- Prototype Declerations  -------------------- */

/* ** Func ConnectServer ************************* */
/*
 * Setup connection to given server and port.
 * Return socket of the connection.
 * On error, return -1.
 */
int ConnectServer( char *servname, unsigned short portnum )
{
/* -- Local Data --------------------------------- */
    struct hostent *servhost;
    struct sockaddr_in server;
    int s;
/* -- Operation ---------------------------------- */
    servhost = gethostbyname( servname );
    if( servhost == NULL ) {
        fprintf( stderr, "gethostbyname.\n" );
        return -1;
    }
    bzero( (char *)&server, sizeof(server) );
    server.sin_family = AF_INET;
    server.sin_port = htons( portnum );
    bcopy( servhost->h_addr, (char *)&server.sin_addr, servhost->h_length );
    if( (s = socket(AF_INET, SOCK_STREAM, 0)) < 0 ) {
        fprintf( stderr, "socket.\n" );
        return -1;
    }
    if( connect( s, (struct sockaddr *)&server, sizeof(server) ) == -1 ) {
        fprintf(stderr, "cannot connect.\n");
        return -1;
    }
    return s;
} /* ConnectServer */

/* ** Func  ********************************** */
/* Show SetDefault. */
void SetDefault( void )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
    port = PORT;
    strcpy( servername, "localhost" );
} /* SetDefault */

/* ** Func Usage ********************************** */
/* Show Usage . */
void Usage( void )
{
/* -- Local Data --------------------------------- */
/* -- Operation ---------------------------------- */
    printf( "Usage : tnrm [switches] notations\n"
            "switches:\n"
            "   -h          display help.\n"
            "   -p port     set port number to connect.\n"
            "   -s server   specify server to connect.\n"
            "notations:\n"
            "   array of single numbers and number ranges\n"
            "   which identifies notations.\n"
            "       single number :  1 3 24\n"
            "       number range  :  1-10 24-27\n"
        );
} /* Usage */

/* ** Func mystrtoi ****************************** */
/*
 * Convert string to integer as decimal.
 * If not number character is included, return -1.
 */
int mystrtoi( char *str )
{
/* -- Local Data --------------------------------- */
    int ret;
    int len;
    int base;
    int i;
    char *p;
/* -- Operation ---------------------------------- */
    for( p = str; *p != '\0'; p++ )
        if( !isdigit( *p ) ) return -1;
    ret = 0;
    for( p = str; *p != '\0'; p++ ) {
        len = strlen( p );
        for( i = 1, base = 1; i < len; i++ ) base *= 10;
        ret += base * ( *p - '0' );
    }
    return ret;
} /* mystrtoi */

/* ** Func ParseArg ********************************** */
/* Parse command line arguments. */
void ParseArg( int argc, char **argv )
{
/* -- Local Data --------------------------------- */
    int i;
    int ExpectPort = FALSE;
    int ExpectServer = FALSE;
/* -- Operation ---------------------------------- */
    for( i = 1; i < argc; i++ ) {
        if( ExpectServer ) {
            strcpy( servername, argv[i] );
            ExpectServer = FALSE;
        } else if( ExpectPort ) {
            port = mystrtoi( argv[i] );
            ExpectPort = FALSE;
        } else if( argv[i][0] == '-' ) {
            switch( argv[i][1] ) {
              case 'h':
                Usage();
                exit( 0 );
              case 'p':
                ExpectPort = TRUE;
                break;
              case 's':
                ExpectServer = TRUE;
                break;
            }
        }
    }
} /* ParseArg */


/* ** Func SendDeleteNotation ********************* */
/* Send DeleteNotation Message */
void SendDeleteNotation( int soc, int id )
{
/* -- Local Data --------------------------------- */
    char buf[BUFSIZE + 1];
/* -- Operation ---------------------------------- */
    sprintf( buf, "DELETENOTATION %d\n", id );
    write( soc, buf, strlen( buf ) );
} /* SendDeleteNotation */

/* ** Func main ********************************** */
/* main func */
int main( int argc, char **argv )
{
/* -- Local Data --------------------------------- */
    int soc;
    const char message[] = "MANAGENOTATION\n";
    int i;
    char *p;
    int id;
    int start;
    int end;
/* -- Operation ---------------------------------- */
    SetDefault();
    ParseArg( argc, argv );
    soc = ConnectServer( servername, port );
    if( soc < 0 ) {
        fprintf( stderr, "cannot connect to server.\n" );
        return 1;
    }
    write( soc, message, strlen( message ) );
    for( i = 1; i < argc; i++ ) {
        if( argv[i][0] == '-' ) continue; /* options */
        p = strtok( argv[i], "-" );
        start = mystrtoi( p );
        if( start < 0 ) continue;
        p = strtok( NULL, "-" );
        if( p == NULL ) {
            SendDeleteNotation( soc, start );
            continue;
        }
        end = mystrtoi( p );
        if( end < 0 ) continue;
        p = strtok( NULL, "-" );
        if( p != NULL ) continue;
        for( id = start; id <= end; id++ )
            SendDeleteNotation( soc, id );
    }
    write( soc, "QUIT\n", 5 );
    close( soc );
    return 0;
} /* main */

